"""
Top level API (:mod:`uv_workon`)
======================================================
"""

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

try:
    __version__ = _version("uv-workon")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "999"


__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "__version__",
]



"""
Core functionality (:mod:`~uv_workon.core`)
===============================================
"""

from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

import attrs

from .validate import (
    infer_virtualenv_name,
    infer_virtualenv_path,
    is_valid_virtualenv,
    validate_dir_exists,
    validate_symlink,
    validate_venv_patterns,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from ._typing import PathLike, VirtualEnvPattern
    from ._typing_compat import Self

logger = logging.getLogger(__name__)


def _converter_pathlike(path: PathLike) -> Path:
    return Path(path)


def _converter_pathlike_absolute(path: PathLike) -> Path:
    return Path(path).absolute()


@attrs.define()
class VirtualEnvPathAndLink:
    """Class to handle virtual environment with link"""

    path: Path = attrs.field(converter=_converter_pathlike)
    link: Path = attrs.field(converter=_converter_pathlike_absolute)

    def create_symlink(
        self,
        resolve: bool = False,
        dry_run: bool = False,
    ) -> None:
        """Create the symlink."""
        path = (
            self.path.resolve()
            if resolve
            else os.path.relpath(self.path.absolute(), self.link.parent)
        )
        logger.info("Creating symlink %s -> %s", self.link, path)
        if not dry_run:
            self.link.unlink(missing_ok=True)
            self.link.symlink_to(path)

    @classmethod
    def from_paths_and_workon(
        cls,
        paths: Iterable[PathLike],
        workon_home: PathLike,
        venv_patterns: VirtualEnvPattern,
        names: str | Iterable[str] | None = None,
    ) -> Iterable[Self]:
        """Get iterable of objects from paths"""
        venv_patterns = validate_venv_patterns(venv_patterns)
        workon_home = validate_dir_exists(workon_home)

        seq: Iterable[tuple[PathLike, str | None]]

        if names is None:
            from itertools import zip_longest

            seq = zip_longest(paths, [names])
        else:
            if isinstance(names, str):
                names = [names]
            seq = zip(paths, names, strict=True)

        for _path, _name in seq:
            if (path := infer_virtualenv_path(_path, venv_patterns)) is not None:
                name = (
                    infer_virtualenv_name(path, venv_patterns)
                    if _name is None
                    else _name
                )
                link = validate_symlink(workon_home / name)
                yield cls(path=path, link=link)


def get_invalid_symlinks(workon_home: Path) -> Iterator[Path]:
    """Get iterator of paths to invalid symlinks under a given path"""
    for path in workon_home.glob("*"):
        if path.is_symlink() and not is_valid_virtualenv(path):
            yield path


def get_virtualenv_paths(
    workon_home: Path,
) -> Iterator[Path]:
    """Get iterator of virtual environment paths under a given path."""
    return (path for path in workon_home.glob("*") if is_valid_virtualenv(path))


def uv_run(
    venv_path: Path,
    *args: str,
    dry_run: bool = False,
) -> str:
    """Construct and run command under uv"""
    import shlex

    args = ("uv", "run", "-p", str(venv_path), "--no-project", *args)
    command: str = (
        f"VIRTUAL_ENV={venv_path} UV_PROJECT_ENVIRONMENT={venv_path} {shlex.join(args)}"
    )

    logger.info("running args: %s", args)
    logger.info("command: %s", command)
    if not dry_run:
        import subprocess

        subprocess.run(
            args,
            check=True,
            env={
                **os.environ,
                "VIRTUAL_ENV": str(venv_path),
                "UV_PROJECT_ENVIRONMENT": str(venv_path),
            },
        )
    return command


def generate_shell_config() -> str:
    """Generate bash/zsh file for shell config."""
    from shutil import which
    from textwrap import dedent

    exe_location = which("uvw")

    return dedent(f"""\
    _UV_WORKON={exe_location}

    _uvw-interface() {{
        local opt="${{2-__missing__}}"
        case "$opt" in
            --help | -h) $_UV_WORKON activate --help ;;
            *) eval $(command $_UV_WORKON $@) ;;
        esac
    }}

    uvw() {{
        local cmd="${{1-__missing__}}"
        case "$cmd" in
            activate | cd) _uvw-interface $@ ;;
            *) command $_UV_WORKON $@ ;;
        esac
    }}

    """)



"""
Interface to type checkers (mypy/pyright) to handle python-version and python-executable.

This allows for running centrally installed (or via uvx) type checkers against a given virtual environment.
"""
# pylint: disable=duplicate-code

from __future__ import annotations

import logging
import os
import shlex
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


FORMAT = "[TYPECHECK %(levelname)s] %(message)s"
logging.basicConfig(level=logging.WARNING, format=FORMAT)
logger = logging.getLogger(__name__)


# * Utilities -----------------------------------------------------------------
def _setup_logging(
    verbosity: int = 0,
) -> None:  # pragma: no cover
    """Setup logging."""
    level_number = max(0, logging.WARNING - 10 * verbosity)
    logger.setLevel(level_number)

    # Silence noisy loggers
    logging.getLogger("sh").setLevel(logging.WARNING)


# * Runner --------------------------------------------------------------------
def _uvx_run(
    *args: str,
    env: Mapping[str, str] | None = None,
    dry_run: bool = False,
) -> None:
    import subprocess

    cleaned_args = ["uvx", *(os.fsdecode(arg) for arg in args)]
    full_cmd = shlex.join(cleaned_args)
    logger.info("Running %s", full_cmd)

    if dry_run:
        return

    r = subprocess.run(cleaned_args, check=False, env=env)
    if returncode := r.returncode:
        logger.error("Command %s failed with exit code %s", full_cmd, returncode)  # pyright: ignore[reportUnknownArgumentType]
        msg = f"Returned code {returncode}"
        raise RuntimeError(msg)


def _run_checker(
    checker: str,
    *args: str,
    python_version: str,
    python_executable: str,
    constraints: list[Path],
    dry_run: bool = False,
) -> None:
    python_flags = (
        (
            f"--pythonpath={python_executable}",
            f"--pythonversion={python_version}",
        )
        if checker == "pyright"
        else (
            f"--python-executable={python_executable}",
            f"--python-version={python_version}",
        )
    )

    _uvx_run(
        *(f"--constraints={c}" for c in constraints),
        checker,
        *python_flags,
        *args,
        dry_run=dry_run,
    )


# * Application ---------------------------------------------------------------
def get_parser() -> ArgumentParser:
    """Get argparser."""
    parser = ArgumentParser(description="Run executable using uvx.")
    parser.add_argument(
        "--python-executable",
        dest="python_executable",
        default=None,  # Path(sys.executable),
        type=Path,
        help="""
        Path to python executable. Defaults to ``sys.executable``. This is
        passed to `--python-executable` in mypy and `--pythonpath` in pyright.
        """,
    )
    parser.add_argument(
        "--python-version",
        dest="python_version",
        default=None,
        type=str,
        help="""
        Python version (x.y) to typecheck against. Defaults to
        ``{sys.version_info.major}.{sys.version_info.minor}``. This is passed
        to ``--python-version`` and ``--pythonversion`` in mypy and pyright.
        """,
    )
    parser.add_argument(
        "-c",
        "--constraints",
        dest="constraints",
        default=[],
        action="append",
        type=Path,
        help="Requirements (requirements.txt) specs for checker.  Can specify multiple times.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbosity",
        action="count",
        default=0,
        help="Set verbosity level.  Pass multiple times to up level.",
    )
    parser.add_argument(
        "-x",
        "--checker",
        dest="checkers",
        default=[],
        action="append",
        choices=["mypy", "pyright"],
        help="Type checker to use.",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
    )
    parser.add_argument(
        "args",
        type=str,
        nargs="*",
        default=[],
        help="extra arguments to checker(s)",
    )

    return parser


def main(args: Sequence[str] | None = None) -> int:
    """Main script."""
    parser = get_parser()
    options = parser.parse_args(args)

    _setup_logging(options.verbosity)

    python_version = (
        options.python_version or f"{sys.version_info.major}.{sys.version_info.minor}"
    )
    python_executable = options.python_executable or sys.executable

    logger.debug("checkers: %s", options.checkers)
    logger.debug("args: %s", options.args)

    for checker in options.checkers:
        _run_checker(
            checker,
            *options.args,
            python_version=python_version,
            python_executable=python_executable,
            constraints=options.constraints,
            dry_run=options.dry_run,
        )

    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from click.testing import CliRunner

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any

    from click import Command
    from pytest_mock import MockerFixture


@pytest.fixture(scope="session")
def click_app() -> Command:
    import uv_workon._click

    return uv_workon._click.click_app  # pylint: disable=protected-access


@pytest.fixture
def example_path(tmp_path: Path) -> Generator[Path]:
    # change to example_path
    old_cwd = Path.cwd()
    os.chdir(tmp_path)

    assert Path.cwd().absolute() == tmp_path.absolute()

    yield tmp_path.absolute()
    # Cleanup?
    os.chdir(old_cwd)

    assert Path.cwd().absolute() == old_cwd.absolute()


@pytest.fixture
def workon_home(example_path: Path) -> Generator[Path]:
    out = example_path / "venvs"
    out.mkdir(exist_ok=True)

    yield out

    shutil.rmtree(out)


@pytest.fixture(scope="session")
def venvs_parent_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    out = tmp_path_factory.mktemp("examples")

    parent_path = out / "a" / "b" / "c"
    parent_path.mkdir(parents=True)

    for i in range(3):
        for args in (
            (f"has_dotvenv_{i}", ".venv"),
            (f"has_venv_{i}", "venv"),
            (f"is_venv_{i}",),
        ):
            d = parent_path.joinpath(*args)
            d.mkdir(parents=True)
            (d / "pyvenv.cfg").write_text("hello")

        for args in (
            (f"bad_dotvenv_{i}", ".venv"),
            (f"bad_venv_{i}", "venv"),
            (f"no_venv_{i}",),
        ):
            d = parent_path.joinpath(*args)
            d.mkdir(parents=True)

    # dummy make activate files ...
    d = parent_path / "is_venv_0" / "bin"
    d.mkdir()
    (d / "activate").touch()

    d = parent_path / "is_venv_1" / "Scripts"
    d.mkdir()
    (d / "activate").touch()

    return parent_path


@pytest.fixture(scope="session")
def clirunner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def workon_home_with_is_venv(
    click_app: Command, workon_home: Path, venvs_parent_path: Path, clirunner: CliRunner
) -> Path:
    paths = venvs_parent_path.glob("is_venv_*")
    clirunner.invoke(
        click_app, ["link", "--workon-home", str(workon_home), "-vv", *map(str, paths)]
    )

    return workon_home


@pytest.fixture
def dummy_kernelspec() -> dict[str, Any]:
    p = Path.cwd().resolve()
    out = {
        name: {
            "resource_dir": str(p / name),
            "spec": {
                "argv": [
                    str(p / name / "bin" / "python"),
                    "-Xfrozen_modules=off",
                    "-m",
                    "ipykernel_launcher",
                    "-f",
                    "{connection_file}",
                ],
                "env": {},
                "display_name": "Python [venv: dummy0]",
                "language": "python",
                "interrupt_mode": "signal",
                "metadata": {"debugger": True},
            },
        }
        for name in ("dummy0", "dummy1")
    }

    out["good"] = {
        "resource_dir": str(p / "good"),
        "spec": {
            "argv": [
                sys.executable,
                "-Xfrozen_modules=off",
                "-m",
                "ipykernel_launcher",
                "-f",
                "{connection_file}",
            ],
            "env": {},
            "display_name": "Python [venv: dummy0]",
            "language": "python",
            "interrupt_mode": "signal",
            "metadata": {"debugger": True},
        },
    }
    return out


@pytest.fixture
def dummy_kernelspec_with_replace(dummy_kernelspec: dict[str, Any]) -> dict[str, Any]:
    p = Path.cwd().resolve()
    name = "is_venv_0"
    return {
        name: {
            "resource_dir": str(p / name),
            "spec": {
                "argv": [
                    str(p / name / "bin" / "python"),
                    "-Xfrozen_modules=off",
                    "-m",
                    "ipykernel_launcher",
                    "-f",
                    "{connection_file}",
                ],
                "env": {},
                "display_name": "Python [venv: dummy0]",
                "language": "python",
                "interrupt_mode": "signal",
                "metadata": {"debugger": True},
            },
        },
        **dummy_kernelspec,
    }


@pytest.fixture
def mocked_get_kernelspec(
    mocker: MockerFixture, dummy_kernelspec: dict[str, Any]
) -> Any:
    return mocker.patch(
        "uv_workon.kernels.get_kernelspecs",
        autospec=True,
        return_value=dummy_kernelspec,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def test_foo(example_path: Path, venvs_parent_path: Path) -> None:
    print(example_path.absolute())
    print(venvs_parent_path.absolute())


def test_bar(example_path: Path, venvs_parent_path: Path) -> None:
    print(example_path.absolute())
    print(venvs_parent_path.absolute())



"""Utilities to wonferrk with nox"""

from __future__ import annotations

import locale
import os
import shlex
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

from nox.logger import logger

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any

    from nox import Session

    PathLike = str | Path


# * Top level installation functions ---------------------------------------------------
def py_prefix(python_version: Any) -> str:
    """
    Get python prefix.

    `python="3.8` -> "py38"
    """
    if isinstance(python_version, str):
        return "py" + python_version.replace(".", "")
    msg = f"passed non-string value {python_version}"
    raise ValueError(msg)


def infer_requirement_path_with_fallback(
    name: str | None,
    ext: str | None = None,
    python_version: str | None = None,
    lock: bool = False,
    check_exists: bool = True,
    lock_fallback: bool = False,
) -> tuple[bool, Path]:
    """Get the requirements file from options with fallback."""
    if lock_fallback:
        try:
            path = infer_requirement_path(
                name=name,
                ext=ext,
                python_version=python_version,
                lock=lock,
                check_exists=True,
            )
        except FileNotFoundError:
            logger.info("Falling back to non-locked")
            lock = False
            path = infer_requirement_path(
                name=name,
                ext=ext,
                python_version=python_version,
                lock=lock,
                check_exists=True,
            )

    else:
        path = infer_requirement_path(
            name=name,
            ext=ext,
            python_version=python_version,
            lock=lock,
            check_exists=check_exists,
        )
    return lock, path


def infer_requirement_path(
    name: str | None,
    ext: str | None = None,
    python_version: str | None = None,
    lock: bool = False,
    check_exists: bool = True,
) -> Path:
    """Get filename for a conda yaml or pip requirements file."""
    if name is None:
        msg = "must supply name"
        raise ValueError(msg)

    # adjust filename
    filename = name
    if ext is not None and not filename.endswith(ext):
        filename += ext
    if python_version is not None and ext != ".txt":
        prefix = py_prefix(python_version)
        if not filename.startswith(prefix):
            filename = f"{prefix}-{filename}"

    if lock:
        if filename.endswith(".yaml"):
            filename = filename.rstrip(".yaml") + "-conda-lock.yml"
        elif filename.endswith(".yml"):
            filename = filename.rstrip(".yml") + "-conda-lock.yml"
        elif filename.endswith(".txt"):
            pass
        else:
            msg = f"unknown file extension for {filename}"
            raise ValueError(msg)

        filename = f"./requirements/lock/{filename}"
    else:
        filename = f"./requirements/{filename}"

    path = Path(filename)
    if check_exists and not path.is_file():
        msg = f"{path} does not exist"
        raise FileNotFoundError(msg)

    return path


def get_python_full_path(session: Session) -> str:
    """Full path to session python executable."""
    path = session.run_always(
        "python",
        "-c",
        "import sys; print(sys.executable)",
        silent=True,
    )
    if not isinstance(path, str):
        msg = "accessing python_full_path with value None"
        raise TypeError(msg)
    return path.strip()


# * Utilities --------------------------------------------------------------------------
def combine_list_str(opts: str | Iterable[str]) -> list[str]:
    """Cleanup str/list[str] to list[str]"""
    if not opts:
        return []

    if isinstance(opts, str):
        opts = [opts]
    return shlex.split(" ".join(opts))


def combine_list_list_str(opts: Iterable[str | Iterable[str]]) -> Iterable[list[str]]:
    """Cleanup Iterable[str/list[str]] to Iterable[list[str]]."""
    return (combine_list_str(opt) for opt in opts)


def open_webpage(path: str | Path | None = None, url: str | None = None) -> None:
    """
    Open webpage from path or url.

    Useful if want to view webpage with javascript, etc., as well as static html.
    """
    import webbrowser
    from urllib.request import pathname2url

    if path:
        url = "file://" + pathname2url(str(Path(path).absolute()))
    if url:
        webbrowser.open(url)


def session_run_commands(
    session: Session,
    commands: list[list[str]] | None,
    external: bool = True,
    **kws: Any,
) -> None:
    """Run commands command."""
    if commands:
        kws.update(external=external)
        for opt in combine_list_list_str(commands):
            session.run(*opt, **kws)


# * Caching -------------------------------------------------------------------
@contextmanager
def check_for_change_manager(
    *deps: str | Path,
    hash_path: str | Path | None = None,
    target_path: str | Path | None = None,
    force_write: bool = False,
) -> Iterator[bool]:
    """
    Context manager to look for changes in dependencies.

    Yields
    ------
    changed: bool

    If exit normally, write hashes to hash_path file

    """
    # pylint: disable=try-except-raise,no-else-raise, too-many-try-statements
    try:
        changed, hashes, hash_path = check_hash_path_for_change(
            *deps,
            target_path=target_path,
            hash_path=hash_path,
        )

        yield changed

    except Exception:  # noqa: TRY203
        raise

    else:
        if force_write or changed:
            logger.info(f"Writing {hash_path}")

            # make sure the parent directory exists:
            hash_path.parent.mkdir(parents=True, exist_ok=True)
            write_hashes(hash_path=hash_path, hashes=hashes)


def check_hash_path_for_change(
    *deps: str | Path,
    target_path: str | Path | None = None,
    hash_path: str | Path | None = None,
) -> tuple[bool, dict[str, str], Path]:
    """
    Checks a json file `hash_path` for hashes of `other_paths`.

    if specify target_path and no hash_path, set `hash_path=target_path.parent / (target_path.name + ".hash.json")`.
    if specify hash_path and no target, set

    Parameters
    ----------
    *deps :
        files on which target_path/hash_path depends.
    hash_path :
        Path of file containing hashes of `deps`.
    target_path :
        Target file (i.e., the final file to be created).
        Defaults to hash_path.


    Returns
    -------
    changed : bool
    hashes : dict[str, str]
    hash_path : Path

    """
    import json

    msg = "Must specify target_path or hash_path"

    if target_path is None:
        if hash_path is None:
            raise ValueError(msg)
        target_path = hash_path = Path(hash_path)
    else:
        target_path = Path(target_path)
        if hash_path is None:
            hash_path = target_path.parent / (target_path.name + ".hash.json")
        else:
            hash_path = Path(hash_path)

    hashes: dict[str, str] = {
        os.path.relpath(k, hash_path.parent): _get_file_hash(k) for k in deps
    }

    if not target_path.is_file():
        changed = True

    elif hash_path.is_file():
        with hash_path.open() as f:
            previous_hashes: dict[str, Any] = json.load(f)

        changed = False
        for k, h in hashes.items():
            previous = previous_hashes.get(k)
            if previous is None or previous != h:
                changed = True
                hashes = {**previous_hashes, **hashes}
                break

    else:
        changed = True

    return changed, hashes, hash_path


def write_hashes(hash_path: str | Path, hashes: dict[str, Any]) -> None:
    """Write hashes to json file."""
    import json

    with Path(hash_path).open("w", encoding=locale.getpreferredencoding(False)) as f:
        json.dump(hashes, f, indent=2)
        f.write("\n")


def _get_file_hash(path: str | Path, buff_size: int = 65536) -> str:
    import hashlib

    md5 = hashlib.md5()  # noqa: S324
    with Path(path).open("rb") as f:
        while data := f.read(buff_size):  # pylint: disable=while-used
            md5.update(data)
    return md5.hexdigest()



from __future__ import annotations

from typing import TYPE_CHECKING

from uv_workon.utils import select_option

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_select_option(mocker: MockerFixture) -> None:
    mock_terminalmenu = mocker.patch("simple_term_menu.TerminalMenu", autospec=True)
    options = ["a", "b"]
    _ = select_option(options, usage=False)
    assert mock_terminalmenu.mock_calls == [
        mocker.call(options, title=None),
        mocker.call().show(),
        mocker.call().show().__index__(),
    ]



"""Script to be called under uv run to install an ipykernel"""

from __future__ import annotations

from argparse import ArgumentParser
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from collections.abc import Sequence

    class _Parser:
        """Dummy class for parser."""

        args: list[str]
        dry_run: bool


def get_parser() -> ArgumentParser:
    """Basic parser."""
    parser = ArgumentParser(description="Interface to python -m ipykernel install")

    parser.add_argument("--dry-run", action="store_true")
    parser.add_argument("args", type=str, nargs="+", default=[])

    return parser


def main(args: Sequence[str] | None = None) -> int:
    """Main program."""
    try:
        import ipykernel  # pyright: ignore[reportUnusedImport] # noqa: F401  # pylint: disable=unused-import
    except ImportError:
        import sys

        print(f"No ipykernel for {sys.executable}")  # noqa: T201
        return 0

    parser = get_parser()
    options = cast(
        "_Parser",
        parser.parse_args() if args is None else parser.parse_args(args),
    )
    args_ = ["python", "-m", "ipykernel", "install", *options.args]

    if options.dry_run:
        print("args", args_)  # noqa: T201
    else:
        import subprocess

        subprocess.run(args_, check=True)

    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main())



from __future__ import annotations

from contextlib import nullcontext
from typing import TYPE_CHECKING

import pytest

from uv_workon.validate import (
    NoVirtualEnvError,
    infer_virtualenv_path_raise,
    is_valid_virtualenv,
    validate_dir_exists,
    validate_is_virtualenv,
    validate_symlink,
    validate_venv_patterns,
)

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from uv_workon._typing import VirtualEnvPattern


@pytest.mark.parametrize(
    ("pattern", "args", "valid"),
    [
        ("has_dotvenv_{}", (".venv",), True),
        ("has_venv_{}", ("venv",), True),
        ("is_venv_{}", (), True),
        ("bad_dotvenv_{}", (".venv",), False),
        ("bad_venv_{}", ("venv",), False),
        ("no_venv_{}", (), False),
    ],
)
def test_is_valid_venv(
    venvs_parent_path: Path, pattern: str, args: tuple[str], valid: bool
) -> None:
    for i in range(3):
        path = venvs_parent_path.joinpath(pattern.format(i), *args)
        print(path)

        assert is_valid_virtualenv(path) == valid

        if valid:
            assert validate_is_virtualenv(path) == path
        else:
            with pytest.raises(
                NoVirtualEnvError, match=r".* is not a valid virtual .*"
            ):
                _ = validate_is_virtualenv(path)


def test_validate_dir_exists(venvs_parent_path: Path) -> None:
    assert validate_dir_exists(venvs_parent_path) == venvs_parent_path

    with pytest.raises(ValueError, match=r".* is not a directory."):
        _ = validate_dir_exists(venvs_parent_path / "a-dummy-dir")


def test_validate_symlink(venvs_parent_path: Path) -> None:
    with pytest.raises(ValueError, match=r".* exists and is not a symlink"):
        _ = validate_symlink(venvs_parent_path)

    p = venvs_parent_path / "tmp"
    p.mkdir()

    link = p / "link"
    link.symlink_to(venvs_parent_path)

    assert validate_symlink(link) == link


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        (("has_dotvenv_0",), nullcontext(("has_dotvenv_0", ".venv"))),
        (("has_venv_0", "venv"), nullcontext(("has_venv_0", "venv"))),
        (("is_venv_0",), nullcontext(("is_venv_0",))),
        (
            ("bad_dotvenv_0",),
            pytest.raises(NoVirtualEnvError, match=r"No venv found .*"),
        ),
        (
            ("bad_dotvenv_0", ".venv"),
            pytest.raises(NoVirtualEnvError, match=r"No venv found .*"),
        ),
        (("no_venv",), pytest.raises(NoVirtualEnvError, match=r"No venv found .*")),
    ],
)
def test_infer_virtualenv_path_raise(
    venvs_parent_path: Path, args: tuple[str], expected: Any
) -> None:
    with expected as e:
        path = venvs_parent_path.joinpath(*args)
        out = infer_virtualenv_path_raise(path, venv_patterns=[".venv", "venv"])

        assert out == venvs_parent_path.joinpath(*e)


@pytest.mark.parametrize(
    ("venv_patterns", "expected"),
    [
        ("hello", ["hello"]),
        (iter(["hello"]), ["hello"]),
        (("hello",), ["hello"]),
        (["hello"], ["hello"]),
        (None, []),
    ],
)
def test_validate_venv_patterns(
    venv_patterns: VirtualEnvPattern, expected: list[str]
) -> None:
    assert validate_venv_patterns(venv_patterns) == expected



"""
Working with ipykernel (:mod:`~uv_workon.kernels`)
==================================================
"""

from __future__ import annotations

import logging
from collections.abc import Iterator  # noqa: TC003
from functools import lru_cache
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


logger = logging.getLogger(__name__)


def has_jupyter_client() -> None:
    """Raise error if does not have jupyter-client"""
    if find_spec("jupyter_client") is None:
        msg = "jupyter_client not installed.  Install it directly or by using `jupyter-client` extra"
        raise ModuleNotFoundError(msg)


def get_ipykernel_install_script_path() -> str:
    """Get the path to ipykernel install script"""
    from importlib.resources import files

    return str(
        files("uv_workon").joinpath("scripts").joinpath("ipykernel_install_script.py")
    )


@lru_cache
def get_kernelspecs() -> dict[str, Any]:
    """Get all kernelspecs"""
    has_jupyter_client()
    from jupyter_client.kernelspecapp import ListKernelSpecs

    return ListKernelSpecs(log_level="ERROR").kernel_spec_manager.get_all_specs()


def get_broken_kernelspecs() -> dict[str, Any]:
    """Get list of broken kernels"""
    from shutil import which

    broken: dict[str, Any] = {}
    for name, data in get_kernelspecs().items():
        exe = data["spec"]["argv"][0]
        if Path(exe).exists() or which(exe):
            continue
        broken[name] = data
    return broken


def remove_kernelspecs(names: list[str]) -> None:
    """Remove named kernels."""
    has_jupyter_client()
    from jupyter_client.kernelspecapp import RemoveKernelSpec

    RemoveKernelSpec(spec_names=names, force=True).start()


def complete_kernelspec_names(incomplete: str) -> Iterator[str]:
    """Complete possible kernel specs"""
    valid_names = get_kernelspecs()
    yield from (name for name in valid_names if name.startswith(incomplete))



"""
Console script (:mod:`~uv_workon.cli`)
======================================
"""

from __future__ import annotations

import itertools
import logging
import sys
from collections.abc import Iterator  # noqa: TC003
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, cast

import click
import typer

from .core import (
    VirtualEnvPathAndLink,
    generate_shell_config,
    get_invalid_symlinks,
    get_virtualenv_paths,
    uv_run,
)
from .kernels import complete_kernelspec_names
from .utils import select_option
from .validate import (
    infer_virtualenv_name,
    infer_virtualenv_path_raise,
    validate_is_virtualenv,
)

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeVar

    R = TypeVar("R")


# * Logging -------------------------------------------------------------------
FORMAT = "[%(name)s - %(levelname)s] %(message)s"
logging.basicConfig(level=logging.WARNING, format=FORMAT)
logger = logging.getLogger(__name__)


# * Utils ---------------------------------------------------------------------
def _get_input_paths(
    paths: list[Path] | None, parents: list[Path] | None
) -> Iterable[Path]:
    if paths is None:
        paths = []
    if parents is None:
        parents = []
    return itertools.chain(paths, *[p.glob("*") for p in parents])


def _select_virtualenv_path(
    venv_path: Path | None,
    venv_name: str | None,
    workon_home: Path,
    venv_patterns: list[str],
    resolve: bool = False,
) -> Path:
    if venv_path:
        path = infer_virtualenv_path_raise(
            venv_path,
            venv_patterns,
        )
    elif venv_name:
        path = validate_is_virtualenv(workon_home / venv_name)

    elif options := [p.name for p in get_virtualenv_paths(workon_home)]:
        path = workon_home / select_option(options, title="venv")
    else:  # pragma: no cover
        typer.echo("No virtual environment found")
        raise typer.Exit(0)

    if resolve:
        path = path.resolve()
    return path


def _get_venv_name_path_mapping(
    include_workon_home: bool,
    venv_names: Iterable[str] | None,
    venv_paths: Iterable[Path] | None,
    workon_home: Path,
    venv_patterns: list[str],
) -> dict[str, Path]:
    name_mapping: dict[str, Path] = {}
    if include_workon_home:
        name_mapping.update({p.name: p for p in get_virtualenv_paths(workon_home)})

    if venv_names:
        name_mapping.update(
            {name: validate_is_virtualenv(workon_home / name) for name in venv_names}
        )

    if venv_paths:
        for p in venv_paths:
            path = infer_virtualenv_path_raise(p, venv_patterns)
            name_mapping[infer_virtualenv_name(path, venv_patterns)] = path

    return name_mapping


def _confirm_action(yes: bool | None, msg: str) -> bool:
    if yes is None:
        return typer.confirm(msg)
    return yes


# * Callbacks -----------------------------------------------------------------
def _callback_expand_user(x: Path) -> Path:
    return x.expanduser()


def _callback_venv_patterns(
    ctx: typer.Context,
    venv_patterns: list[str],
) -> list[str]:
    use_default = cast("bool", ctx.params.get("use_default_venv_patterns"))
    return list({*venv_patterns, *((".venv", "venv") if use_default else ())})


def _callback_verbose(
    verbose: int | None,
) -> int | None:
    if verbose is None:
        return verbose

    if verbose < 0:  # pragma: no cover
        level = logging.ERROR
    elif not verbose:  # pragma: no cover
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    else:  # pragma: no cover
        level = logging.DEBUG

    for _logger in map(logging.getLogger, logging.root.manager.loggerDict):  # pylint: disable=no-member
        _logger.setLevel(level)
    return verbose


# * Completions ---------------------------------------------------------------
@lru_cache
def _all_virtualenv_names(workon_home: Path) -> list[str]:
    return [p.name for p in get_virtualenv_paths(workon_home)]


def _complete_virtualenv_names(ctx: typer.Context, incomplete: str) -> Iterator[str]:
    workon_home = ctx.params.get("workon_home")
    valid_names = _all_virtualenv_names(workon_home)
    yield from (name for name in valid_names if name.startswith(incomplete))


# this is necessary because typer has a bug where any path typed is considered
# 'complete' for arguments (even if exists=True) and a space is added to the
# end - this works around it.
def _complete_path() -> list[str]:
    return []


# * Main app ------------------------------------------------------------------
app_typer = typer.Typer(no_args_is_help=True)
app_kernels = typer.Typer(no_args_is_help=True, help="Jupyter kernel utilities")
app_typer.add_typer(app_kernels, name="kernels")


def version_callback(value: bool) -> None:
    """Versioning call back."""
    from uv_workon import __version__

    if value:
        typer.echo(f"uv-workon, version {__version__}")
        raise typer.Exit


@app_typer.callback()
def main(
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, is_eager=True
    ),
) -> None:
    """Manage uv virtual environments from central location."""
    return


# * Options -------------------------------------------------------------------
WORKON_HOME_CLI = Annotated[
    Path,
    typer.Option(
        "--workon-home",
        "-o",
        help="""
        Directory containing the virtual environments and links to virtual
        environments. If not passed, uses in order, ``WORKON_HOME`` environment
        variable, then ``~/.virtualenvs`` directory.
        """,
        envvar="WORKON_HOME",
        callback=_callback_expand_user,
        is_eager=True,  # needed for autocompletion
        autocompletion=_complete_path,
        default_factory=lambda: Path.home() / ".virtualenvs",
        show_default=False,
    ),
]
DRY_RUN_CLI = Annotated[
    bool,
    typer.Option(
        "--dry-run/--no-dry-run",
        help="Perform a dry run, without executing any action",
    ),
]
VERBOSE_CLI = Annotated[
    int | None,
    typer.Option(
        "--verbose",
        "-v",
        help="Set verbosity level.  Can specify multiple times",
        count=True,
        callback=_callback_verbose,
    ),
]
VENV_PATTERNS_CLI = Annotated[
    list[str],
    typer.Option(
        "--venv",
        help="""
        Virtual environment pattern. Can specify multiple times.
        Default is to include virtual environment directories of form
        ``".venv"`` or ``"venv"``.  To exclude these defaults, pass ``--no-default-venv``.
        """,
        envvar="UV_WORKON_VENV_PATTERNS",
        callback=_callback_venv_patterns,
        default_factory=list,
        show_default=False,
    ),
]
USE_DEFAULT_VENV_PATTERNS_CLI = Annotated[
    bool,
    typer.Option(
        "--default-venv/--no-default-venv",
        help="""
        Default is to include virtual environment patterns ``".venv"`` and ``"venv"``.
        Pass ``--no-default-venv`` to exclude these default values.
        """,
        is_eager=True,
    ),
]
RESOLVE_CLI = Annotated[
    bool,
    typer.Option(
        "--resolve/--no-resolve",
        help="""
        Pass ``--resolve`` to resolve paths and symlinks.  Otherwise, use relative paths.
        """,
    ),
]
PATHS_CLI = Annotated[
    list[Path] | None,
    typer.Argument(
        help="""
        Paths to virtual environments. These can either be full paths to
        virtual environments, or path to the parent of a virtual environment
        that has name ``venv_pattern``. If the name (the last element) of the
        path matches ``venv_pattern``, then the name of the linked virtual
        environment will come from the parent directory. Otherwise, it will be
        the name.
        """,
        autocompletion=_complete_path,
        envvar="UV_WORKON_PATHS",
    ),
]
LINK_NAMES_CLI = Annotated[
    list[str] | None,
    typer.Option(
        "--link-name",
        help="""
        Name of the linked virtual environment. Default is to infer the name
        from path. Can specify multiple times. If use this option, it must
        match up with the number of paths. It is intended to be used once only.
        Use with care in other cases.
        """,
    ),
]
PARENTS_CLI = Annotated[
    list[Path] | None,
    typer.Option(
        "--parent",
        help="""
        Parent of directories to check for ``venv_pattern`` directories
        containing virtual environments. Using ``uv-workon --parent a/path``
        is roughly equivalent to using ``uv-workon a/path/*``
        """,
        autocompletion=_complete_path,
    ),
]
YES_CLI = Annotated[
    bool | None,
    typer.Option(
        "--yes/--no",
        help="""
        Default is confirm yes/No for actions.
        Passing ``--yes`` confirms all answers as yes.
        Passing ``--no`` confirms all answers as no.
        """,
    ),
]
VENV_NAME_CLI = Annotated[
    str | None,
    typer.Option(
        "--name",
        "-n",
        help="Use virtual environment located at ${workon_home}/{name}.",
        autocompletion=_complete_virtualenv_names,
    ),
]
VENV_PATH_CLI = Annotated[
    Path | None,
    typer.Option(
        "-p",
        "--path",
        help="""Path to venv""",
        autocompletion=_complete_path,
    ),
]
UV_RUN_OPTIONS_CLI = Annotated[
    list[str], typer.Argument(help="Arguments and options passed to ``uv run ...``")
]
NO_COMMAND_CLI = Annotated[
    bool,
    typer.Option(
        "--command/--no-command",
        help="""
        If ``--command``, include command name with output.
        If ``--no-command``, only list the path.
        """,
    ),
]
# Multiple names/paths
VENV_NAMES_CLI = Annotated[
    list[str] | None,
    typer.Option(
        "-n",
        "--name",
        help="Virtual environment names.",
        autocompletion=_complete_virtualenv_names,
    ),
]
VENV_PATHS_CLI = Annotated[
    list[Path] | None,
    typer.Option(
        "-p",
        "--path",
        help="Virtual environment paths.",
        autocompletion=_complete_path,
    ),
]


# * Commands ------------------------------------------------------------------
# ** Links
@app_typer.command("link")
def link_virtualenvs(
    *,
    paths: PATHS_CLI = None,
    parents: PARENTS_CLI = None,
    link_names: LINK_NAMES_CLI = None,
    resolve: RESOLVE_CLI = False,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
    dry_run: DRY_RUN_CLI = False,
    verbose: VERBOSE_CLI = None,
    yes: YES_CLI = None,
) -> None:
    """Create symlink from paths to workon_home."""
    if not (input_paths := list(_get_input_paths(paths, parents))):
        with click.get_current_context() as ctx:
            typer.echo(ctx.get_help())
            sys.exit(2)

    logger.debug("params: %s", locals())

    objs = list(
        VirtualEnvPathAndLink.from_paths_and_workon(
            input_paths,
            workon_home=workon_home,
            venv_patterns=venv_patterns,
            names=link_names,
        )
    )

    for obj in objs:
        if (not obj.link.exists()) or (
            obj.link.is_symlink() and _confirm_action(yes, f"Overwrite {obj.link}")
        ):
            obj.create_symlink(resolve=resolve, dry_run=dry_run)
        else:
            logger.debug("Skipping: %s -> %s", obj.link, obj.path)


@app_typer.command("list")
def list_virtualenvs(
    *,
    workon_home: WORKON_HOME_CLI,
    verbose: VERBOSE_CLI = None,
) -> None:
    """List available central virtual environments"""
    venv_paths = get_virtualenv_paths(workon_home)
    logger.debug("params: %s", locals())

    for p in sorted(venv_paths, key=lambda x: x.name):
        typer.echo(f"{p.name:25}  {p.resolve()}")


@app_typer.command("clean")
def clean_virtualenvs(
    *,
    workon_home: WORKON_HOME_CLI,
    dry_run: DRY_RUN_CLI = False,
    verbose: VERBOSE_CLI = None,
    yes: YES_CLI = None,
) -> None:
    """Remove missing broken virtual environment symlinks."""
    logger.debug("params: %s", locals())

    for path in get_invalid_symlinks(workon_home):
        if _confirm_action(yes, f"Remove {path} -> {path.readlink()}"):
            logger.info("Remove symlink: %s -> %s", path, path.readlink())
            if not dry_run:
                path.unlink()


@app_typer.command(
    "run",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def run_with_virtualenv(
    *,
    ctx: typer.Context,
    venv_name: VENV_NAME_CLI = None,
    venv_path: VENV_PATH_CLI = None,
    resolve: RESOLVE_CLI = True,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
    dry_run: DRY_RUN_CLI = False,
    verbose: VERBOSE_CLI = None,
) -> None:
    """
    Run uv commands using using the named or specified virtual environment.

    For example, use ``uvw run -n my-env -- python ...`` is
    translated to ``uv run -p patt/to/my-env --no-project python ...``.
    If an option mirrors one of the command options (-n, etc), pass it after ``--``.
    Use ``--dry-run`` to echo the equivalent command to be run in the shell.

    Under the hood, this uses ``uv run`` with the environment variables
    ``VIRTUAL_ENV`` and ``UV_PROJECT_ENVIRONMENT`` and the option ``-p`` all
    set to the path of the virtual environment, and with the option
    ``--no-project``.
    """
    logger.info("params: %s", locals())

    if not ctx.args:
        typer.echo(ctx.get_help())
        sys.exit(2)

    path = _select_virtualenv_path(
        venv_path=venv_path,
        venv_name=venv_name,
        workon_home=workon_home,
        venv_patterns=venv_patterns,
        resolve=resolve,
    )

    command = uv_run(path, *ctx.args, dry_run=dry_run)
    if dry_run:  # pragma: no branch
        typer.echo(command)


# ** Shell commands
@app_typer.command("shell-config")
def shell_config() -> None:
    """
    Use with ``eval "$(uv-workon shell-config)"``.


    This will add the subcommand ``uvw activate`` and ``uvw cd`` to the shell.  Without
    running shell config, ``activate`` and ``cd`` will just print the command to screen.
    """
    typer.echo(generate_shell_config())


@app_typer.command("activate")
def shell_activate(
    *,
    venv_name: VENV_NAME_CLI = None,
    venv_path: VENV_PATH_CLI = None,
    resolve: RESOLVE_CLI = False,
    no_command: NO_COMMAND_CLI = False,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
) -> None:
    """Use to activate virtual environments."""
    path = _select_virtualenv_path(
        venv_path=venv_path,
        venv_name=venv_name,
        workon_home=workon_home,
        venv_patterns=venv_patterns,
        resolve=resolve,
    )

    if (activate := path / "bin" / "activate").exists() or (
        activate := path / "Scripts" / "activate"
    ).exists():
        typer.echo(str(activate) if no_command else f"source {activate}")
    else:
        logger.error("No activate script found for path %s", path)
        raise typer.Exit(1)


@app_typer.command("cd")
def shell_cd(
    *,
    venv_name: VENV_NAME_CLI = None,
    venv_path: VENV_PATH_CLI = None,
    no_command: NO_COMMAND_CLI = False,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
) -> None:
    """Command to change to parent directory of virtual environment."""
    path = _select_virtualenv_path(
        venv_path=venv_path,
        venv_name=venv_name,
        workon_home=workon_home,
        venv_patterns=venv_patterns,
        resolve=True,
    ).parent

    typer.echo(str(path) if no_command else f"cd {path}")


# ** Kernels
@app_kernels.command("install")
def install_ipykernels(
    ctx: typer.Context,
    *,
    venv_names: VENV_NAMES_CLI = None,
    venv_paths: VENV_PATHS_CLI = None,
    all_venvs: Annotated[
        bool, typer.Option("--all", help="If passed, install for all environments")
    ] = False,
    display_format: Annotated[
        str,
        typer.Option(
            "--display-format",
            help="""
            Format to use in display name. Can contain ``name`` which is
            inferred from the virtual environment location.
            """,
        ),
    ] = "Python [venv: {name}]",
    no_user: Annotated[
        bool,
        typer.Option(
            "--user/--no-user",
            help="Default is to pass the ``--user`` options.  Use this to override.",
        ),
    ] = False,
    resolve: RESOLVE_CLI = True,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
    dry_run: DRY_RUN_CLI = False,
    verbose: VERBOSE_CLI = None,
    yes: YES_CLI = None,
) -> None:
    """Install ipykernels for virtual environment(s) that contain ``ipykernel`` module."""
    from .kernels import get_ipykernel_install_script_path, get_kernelspecs

    script = get_ipykernel_install_script_path()
    kernelspecs = get_kernelspecs()

    for name, path in _get_venv_name_path_mapping(
        all_venvs,
        venv_names=venv_names,
        venv_paths=venv_paths,
        workon_home=workon_home,
        venv_patterns=venv_patterns,
    ).items():
        if (name not in kernelspecs) or _confirm_action(yes, f"Reinstall {name}?"):
            display_name = display_format.format(name=name)
            command = uv_run(
                path.resolve() if resolve else path,
                "python",
                script,
                *(["--dry-run"] if dry_run else []),
                "--",
                *ctx.args,
                "--name",
                name,
                "--display-name",
                display_name,
                *([] if no_user else ["--user"]),
                dry_run=dry_run,
            )

            if dry_run:
                typer.echo(command)


@app_kernels.command("remove")
def remove_kernels(
    *,
    names: Annotated[
        list[str] | None,
        typer.Option(
            "-n",
            "--name",
            help="Kernel names to remove",
            autocompletion=complete_kernelspec_names,
        ),
    ] = None,
    venv_paths: VENV_PATHS_CLI = None,
    missing: Annotated[
        bool,
        typer.Option("--missing", help="Remove kernelspecs that are missing/broken."),
    ] = False,
    workon_home: WORKON_HOME_CLI,
    venv_patterns: VENV_PATTERNS_CLI,
    use_default_venv_patterns: USE_DEFAULT_VENV_PATTERNS_CLI = True,
    dry_run: DRY_RUN_CLI = False,
    verbose: VERBOSE_CLI = None,
    yes: YES_CLI = None,
) -> None:
    """Remove installed kernels"""
    from .kernels import (
        get_broken_kernelspecs,
        get_kernelspecs,
        has_jupyter_client,
        remove_kernelspecs,
    )

    has_jupyter_client()

    to_remove = set(
        _get_venv_name_path_mapping(
            include_workon_home=False,
            venv_names=None,
            venv_paths=venv_paths,
            workon_home=workon_home,
            venv_patterns=venv_patterns,
        )
    )
    if names is not None:
        to_remove.update(set(names))

    if missing:
        to_remove.update(set(get_broken_kernelspecs()))

    to_remove = to_remove.intersection(get_kernelspecs())

    to_remove_filtered = sorted(
        name for name in to_remove if _confirm_action(yes, f"Remove {name}")
    )

    if not to_remove_filtered:
        return

    if not dry_run:
        remove_kernelspecs(to_remove_filtered)


@app_kernels.command("list")
def list_kernels() -> None:
    """List installed kernels.  Interface to jupyter kernelspec list."""
    from .kernels import has_jupyter_client

    has_jupyter_client()

    from jupyter_client.kernelspecapp import ListKernelSpecs

    ListKernelSpecs(log_level="ERROR").start()



"""
Utilities (:mod:`~uv_workon.utils`)
===================================
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


def select_option(
    options: Sequence[str],
    title: str = "",
    usage: bool = True,
) -> str:
    """Use selector"""
    from simple_term_menu import (  # pyright: ignore[reportMissingTypeStubs]
        TerminalMenu,
    )

    title = " ".join(
        [
            *([title] if title else []),
            *(
                ["use arrows or j/k to move down/up, or / to limit by name"]
                if usage
                else []
            ),
        ]
    )
    index: int = TerminalMenu(options, title=title or None).show()  # pyright: ignore[reportAssignmentType]
    return options[index]



from __future__ import annotations

import os
from functools import partial
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING

import pytest

from uv_workon.core import (
    VirtualEnvPathAndLink,
    generate_shell_config,
    uv_run,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def find_venvs_interface(
    *args: Path, workon_home: Path, venv_patterns: list[str] | None = None
) -> list[Path]:
    if venv_patterns is None:
        venv_patterns = [".venv", "venv"]

    return [
        obj.path
        for obj in VirtualEnvPathAndLink.from_paths_and_workon(
            args, workon_home=workon_home, venv_patterns=venv_patterns
        )
    ]


def test_find_venvs_explicit(venvs_parent_path: Path, workon_home: Path) -> None:
    find_venvs = partial(find_venvs_interface, workon_home=workon_home)
    paths_dotvenv = list(venvs_parent_path.glob("has_dotvenv_*/.venv"))

    assert find_venvs(*paths_dotvenv) == paths_dotvenv
    assert find_venvs(*venvs_parent_path.glob("has_dotvenv_*")) == paths_dotvenv
    assert find_venvs(*venvs_parent_path.glob("has_dotvenv_*"), venv_patterns=[]) == []

    paths_venv = list(venvs_parent_path.glob("has_venv_*/venv"))
    assert find_venvs(*paths_venv) == paths_venv

    paths_is_venv = list(venvs_parent_path.glob("is_venv_*"))
    assert find_venvs(*paths_is_venv) == paths_is_venv

    assert find_venvs(*venvs_parent_path.glob("bad_dotvenv_*/.venv")) == []
    assert find_venvs(*venvs_parent_path.glob("bad_dotvenv_*")) == []

    assert find_venvs(*venvs_parent_path.glob("bad_venv_*/venv")) == []
    assert find_venvs(*venvs_parent_path.glob("bad_venv_*")) == []

    assert find_venvs(*venvs_parent_path.glob("no_venv_*")) == []


def test_find_venvs_with_name(venvs_parent_path: Path, workon_home: Path) -> None:
    paths = sorted(venvs_parent_path.glob("has_dotvenv_*"), key=lambda x: x.name)

    out = list(
        VirtualEnvPathAndLink.from_paths_and_workon(
            paths, workon_home=workon_home, venv_patterns=[".venv", "venv"]
        )
    )

    assert [x.link.name for x in out] == [f"has_dotvenv_{i}" for i in range(3)]

    out = list(
        VirtualEnvPathAndLink.from_paths_and_workon(
            paths,
            names=["a", "b", "c"],
            workon_home=workon_home,
            venv_patterns=[".venv", "venv"],
        )
    )

    assert [x.link.name for x in out] == ["a", "b", "c"]

    out = list(
        VirtualEnvPathAndLink.from_paths_and_workon(
            paths[:1],
            names="a",
            workon_home=workon_home,
            venv_patterns=[".venv", "venv"],
        )
    )

    assert [x.link.name for x in out] == ["a"]

    with pytest.raises(ValueError, match=r".* shorter.*"):
        out = list(
            VirtualEnvPathAndLink.from_paths_and_workon(
                paths,
                names="a",
                workon_home=workon_home,
                venv_patterns=[".venv", "venv"],
            )
        )


def test_generate_shell_config() -> None:
    assert "_UV_WORKON" in generate_shell_config()


def test_uv_run_error() -> None:
    from tempfile import TemporaryDirectory

    args = ["python", "-c", "import sys; print(sys.executable)"]
    with TemporaryDirectory() as e, pytest.raises(CalledProcessError):
        _ = uv_run(Path(e), *args, dry_run=False)

    import sys

    _ = uv_run(Path(sys.executable), *args, dry_run=False)


def test_uv_run(mocker: MockerFixture) -> None:
    mock_subprocess_run = mocker.patch("subprocess.run", autospec=True)

    venv_path = Path.cwd().resolve()
    args = ["python", "-c", "import sys"]

    uv_run(venv_path, *args)

    # pylint: disable=duplicate-code
    assert mock_subprocess_run.mock_calls == [
        mocker.call(
            (
                "uv",
                "run",
                "-p",
                str(venv_path),
                "--no-project",
                *args,
            ),
            check=True,
            env={
                **os.environ,
                "VIRTUAL_ENV": str(venv_path),
                "UV_PROJECT_ENVIRONMENT": str(venv_path),
            },
        )
    ]



"""
Validation/inference (:mod:`~uv_workon.validate`)
=================================================
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Container, Iterable

    from ._typing import PathLike, VirtualEnvPattern


class NoVirtualEnvError(ValueError):
    """Error to raise if no virtual environment found."""


# ** Validate
def validate_venv_patterns(venv_patterns: VirtualEnvPattern) -> list[str]:
    """Validate venv patterns."""
    # fast exit for most likely cas
    if isinstance(venv_patterns, list):
        return venv_patterns

    if venv_patterns is None:
        return []

    if isinstance(venv_patterns, str):
        return [venv_patterns]

    return list(venv_patterns)


def is_valid_virtualenv(path: Path) -> bool:
    """Check if path is a valid venv"""
    return path.is_dir() and (path / "pyvenv.cfg").exists()


def validate_is_virtualenv(path: PathLike) -> Path:
    """Validate is a virtual environment path"""
    path = Path(path)
    if not is_valid_virtualenv(path):
        msg = f"{path} is not a valid virtual environment"
        raise NoVirtualEnvError(msg)
    return path


def validate_dir_exists(path: PathLike) -> Path:
    """Validate that path is a directory."""
    path = Path(path)
    if not path.is_dir():
        msg = f"{path} is not a directory."
        raise ValueError(msg)
    return path


def validate_symlink(path: PathLike) -> Path:
    """If path exists, assert it is a symlink"""
    path = Path(path)
    if path.exists() and not path.is_symlink():
        msg = f"{path} exists and is not a symlink"
        raise ValueError(msg)
    return path


# ** Infer
def infer_virtualenv_name(path: Path, venv_patterns: Container[str]) -> str:
    """Infer a virtual environment name from path."""
    path_resolved = path.resolve()
    if (name := path_resolved.name) in venv_patterns:
        return path_resolved.parent.name
    return name


def infer_virtualenv_path(
    path: PathLike,
    venv_patterns: Iterable[str],
) -> Path | None:
    """Find a virtual env by pattern and return None if not found."""
    path = Path(path)
    if is_valid_virtualenv(path):
        return path
    for pattern in venv_patterns:
        if is_valid_virtualenv(path_pattern := path / pattern):
            return path_pattern
    return None


def infer_virtualenv_path_raise(
    path: PathLike,
    venv_patterns: Iterable[str],
) -> Path:
    """Find a virtual env by pattern and raise if not found."""
    if (path_ := infer_virtualenv_path(path, venv_patterns)) is None:
        msg = f"No venv found at {path}"
        raise NoVirtualEnvError(msg)
    return path_



# pyright: reportPrivateUsage=false
# pylint: disable=protected-access
from __future__ import annotations

import os
import shlex
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
import typer

from uv_workon import cli
from uv_workon.core import generate_shell_config
from uv_workon.kernels import get_ipykernel_install_script_path

from .test_kernels import skip_if_no_jupyter_client

if TYPE_CHECKING:
    from typing import Any

    from click import Command
    from click.testing import CliRunner
    from pytest_mock import MockerFixture
    from typer import Context


# * Callbacks
def test__get_input_paths(venvs_parent_path: Path) -> None:
    paths = list(venvs_parent_path.glob("has_dotvenv_*/.venv"))
    out = list(cli._get_input_paths(paths, parents=[]))

    expected = {
        venvs_parent_path / f"{x}_{i}" / y
        for x, y in (("has_dotvenv", ".venv"),)
        for i in range(3)
    }

    assert set(out) == expected

    # adding in a venv
    out = list(
        cli._get_input_paths(
            paths,
            parents=[venvs_parent_path / "has_venv_0"],
        )
    )

    expected2 = expected.copy()
    expected2.add(venvs_parent_path / "has_venv_0" / "venv")

    assert set(out) == expected2


def test__select_venv_path(
    venvs_parent_path: Path,
    workon_home_with_is_venv: Path,
    mocker: MockerFixture,
) -> None:
    mock_terminalmenu = mocker.patch("simple_term_menu.TerminalMenu", autospec=True)
    func = partial(
        cli._select_virtualenv_path,
        workon_home=workon_home_with_is_venv,
        venv_patterns=[".venv", "venv"],
    )

    path = func(venv_path=None, venv_name="is_venv_0")
    assert path == workon_home_with_is_venv / "is_venv_0"

    path = func(venv_path=venvs_parent_path / "is_venv_0", venv_name=None)
    assert path == venvs_parent_path / "is_venv_0"

    path = func(venv_path=None, venv_name=None)

    options = [p.name for p in workon_home_with_is_venv.glob("*")]
    assert mock_terminalmenu.mock_calls == [
        mocker.call(
            options,
            title="venv use arrows or j/k to move down/up, or / to limit by name",
        ),
        mocker.call().show(),
        mocker.call().show().__index__(),
    ]


def test__get_venv_name_path_mapping(
    venvs_parent_path: Path,
    workon_home_with_is_venv: Path,
) -> None:
    func = partial(
        cli._get_venv_name_path_mapping,
        workon_home=workon_home_with_is_venv,
        venv_patterns=[".venv", "venv"],
    )

    mapping = func(True, None, None)

    expected = {
        name: workon_home_with_is_venv / name
        for name in (f"is_venv_{i}" for i in range(3))
    }
    assert mapping == expected

    mapping = func(False, ["is_venv_0"], [venvs_parent_path / "has_dotvenv_0"])
    assert mapping == {
        "is_venv_0": workon_home_with_is_venv / "is_venv_0",
        "has_dotvenv_0": venvs_parent_path / "has_dotvenv_0" / ".venv",
    }


@pytest.mark.parametrize(
    ("yes", "mock_value", "expected"),
    [
        (None, False, False),
        (None, True, True),
        (False, False, False),
        (True, False, True),
        (False, True, False),
        (True, True, True),
    ],
)
def test__confirm_action(
    mocker: MockerFixture, yes: bool | None, mock_value: bool, expected: bool
) -> None:
    mocker.patch("typer.confirm", autospec=True, return_value=mock_value)
    func = cli._confirm_action

    assert func(yes, "hello") is expected


def test__add_verbose_logger() -> None:
    import logging

    func = cli._callback_verbose

    func(verbose=-1)
    assert cli.logger.level == logging.ERROR

    func(verbose=0)
    assert cli.logger.level == logging.WARNING

    func(verbose=1)
    assert cli.logger.level == logging.INFO

    func(verbose=2)
    assert cli.logger.level == logging.DEBUG

    # no change with None
    func(verbose=None)
    assert cli.logger.level == logging.DEBUG


# * Callbacks/defaults
@pytest.fixture(scope="session")
def workon_home_click_app() -> Command:
    app = typer.Typer()

    @app.command()
    def dummy(workon_home: cli.WORKON_HOME_CLI) -> Path:  # pyright: ignore[reportUnusedFunction]
        typer.echo(str(workon_home))
        return workon_home

    return typer.main.get_command(app)


@pytest.mark.parametrize(
    ("cli_val", "environment_val", "expected"),
    [
        ("~/hello", None, "~/hello"),
        (None, "~/there", "~/there"),
        (None, None, "~/.virtualenvs"),
        ("~/a", "~/b", "~/a"),
        ("a/b", None, "a/b"),
    ],
)
def test_workon_home_option(
    workon_home_click_app: Command,
    clirunner: CliRunner,
    environment_val: str | None,
    cli_val: str | None,
    expected: Path,
) -> None:
    env = {"WORKON_HOME": environment_val} if environment_val else {}
    opts = [] if cli_val is None else ["--workon-home", cli_val]
    out = clirunner.invoke(workon_home_click_app, opts, env=env)
    assert not out.exit_code
    assert str(Path(expected).expanduser()) == out.output.strip()


@pytest.fixture(scope="session")
def venv_patterns_app() -> Command:
    app = typer.Typer()

    @app.command()
    def dummy(  # pyright: ignore[reportUnusedFunction]
        *,
        venv_patterns: cli.VENV_PATTERNS_CLI,
        use_default_venv_patterns: cli.USE_DEFAULT_VENV_PATTERNS_CLI = True,  # noqa: ARG001
    ) -> None:
        assert isinstance(venv_patterns, list)
        typer.echo(str(sorted(venv_patterns)))

    return typer.main.get_command(app)


@pytest.mark.parametrize(
    ("environment_val", "venv_patterns", "use_default", "expected"),
    [  # pyright: ignore[reportUnknownArgumentType]
        (None, [], True, {".venv", "venv"}),
        (None, ["hello"], True, {"hello", ".venv", "venv"}),
        (None, ["hello", "there"], True, {"hello", "there", ".venv", "venv"}),
        ("a", [], True, {"a", ".venv", "venv"}),
        ("a b", [], True, {"a", "b", ".venv", "venv"}),
        ("a b", ["hello"], True, {"hello", ".venv", "venv"}),
        (None, [], False, {}),
        (None, ["hello"], False, {"hello"}),
        ("a", [], False, {"a"}),
    ],
)
def test_venv_patterns_option(
    venv_patterns_app: Command,
    clirunner: CliRunner,
    environment_val: str | None,
    venv_patterns: list[str],
    use_default: bool,
    expected: set[str],
) -> None:
    env = {"UV_WORKON_VENV_PATTERNS": environment_val} if environment_val else {}
    out = clirunner.invoke(
        venv_patterns_app,
        [
            *[f"--venv={x}" for x in venv_patterns],
            ("--default-venv" if use_default else "--no-default-venv"),
        ],
        env=env,
    )

    assert not out.exit_code
    assert str(sorted(expected)) == out.output.strip()


# * Completion
def test__complete_path() -> None:
    assert cli._complete_path() == []  # pylint: disable=use-implicit-booleaness-not-comparison


# * Version
def test_version(
    click_app: Command,
    clirunner: CliRunner,
) -> None:
    from uv_workon import __version__

    out = clirunner.invoke(
        click_app,
        ["--version"],
    )

    assert f"uv-workon, version {__version__}" in out.output


# * Commands
@pytest.mark.parametrize(
    ("pattern", "parts"),
    [
        ("is_venv*", ("is_venv_{i}",)),
        ("has_dotvenv_*", ("has_dotvenv_{i}", ".venv")),
        ("has_dotvenv_*/.venv", ("has_dotvenv_{i}", ".venv")),
        ("has_venv_*", ("has_venv_{i}", "venv")),
        ("has_venv_*/venv", ("has_venv_{i}", "venv")),
        ("bad_dotvenv_*", None),
        ("bad_dotvenv_*/.venv", None),
    ],
)
@pytest.mark.parametrize("resolve", [True, False])
@pytest.mark.parametrize("dry", [True, False])
def test_link_paths(
    click_app: Command,
    clirunner: CliRunner,
    workon_home: Path,
    venvs_parent_path: Path,
    pattern: str,
    parts: tuple[str] | None,
    resolve: bool,
    dry: bool,
) -> None:
    paths = venvs_parent_path.glob(pattern)

    clirunner.invoke(
        click_app,
        [
            "link",
            "--workon-home",
            str(workon_home),
            "-vv",
            *map(str, paths),
            *(["--resolve"] if resolve else []),
            *(["--dry-run"] if dry else []),
        ],
    )

    if dry:
        assert set(workon_home.glob("*")) == set()
        return

    expected_symlinks: set[Path] = (
        set()
        if parts is None
        else {workon_home / parts[0].format(i=i) for i in range(3)}
    )

    assert expected_symlinks == set(workon_home.glob("*"))

    if parts is not None:
        # look at readlink
        expected_paths = {
            venvs_parent_path.joinpath(*[_.format(i=i) for _ in parts])
            for i in range(3)
        }

        if not resolve:
            expected_paths = {
                Path(os.path.relpath(p, workon_home)) for p in expected_paths
            }

        print(expected_paths)
        assert expected_paths == {p.readlink() for p in workon_home.glob("*")}


def test_link_parent(
    click_app: Command, clirunner: CliRunner, workon_home: Path, venvs_parent_path: Path
) -> None:
    clirunner.invoke(
        click_app,
        [
            "link",
            "--workon-home",
            str(workon_home),
            "-vv",
            "--parent",
            str(venvs_parent_path),
        ],
    )

    expected_symlinks = {
        workon_home / fmt.format(i=i)
        for fmt in ("has_dotvenv_{i}", "has_venv_{i}", "is_venv_{i}")
        for i in range(3)
    }

    assert expected_symlinks == set(workon_home.glob("*"))


def test_link_help(
    click_app: Command,
    clirunner: CliRunner,
) -> None:
    out = clirunner.invoke(click_app, ["link"])

    assert "Create symlink from paths" in out.output


def test_list(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
) -> None:
    out = clirunner.invoke(
        click_app, ["list", "--workon-home", str(workon_home_with_is_venv)]
    )

    links = sorted(workon_home_with_is_venv.glob("*"), key=lambda x: x.name)
    expected = "\n".join([f"{p.name:25}  {p.resolve()}" for p in links])
    assert expected == out.output.strip()


def test_name_completions(
    workon_home_with_is_venv: Path,
) -> None:
    class Dummy:
        """Dummy class"""

        def __init__(self, workon_home: Path) -> None:
            self.params: dict[str, Any] = {"workon_home": workon_home}

    d = cast("Context", Dummy(workon_home=workon_home_with_is_venv))
    assert sorted(cli._complete_virtualenv_names(d, "")) == [
        f"is_venv_{i}" for i in range(3)
    ]
    assert sorted(cli._complete_virtualenv_names(d, "is_venv_")) == [
        f"is_venv_{i}" for i in range(3)
    ]
    assert sorted(cli._complete_virtualenv_names(d, "is_venv_0")) == [
        f"is_venv_{i}" for i in range(1)
    ]


@pytest.mark.parametrize("dry", [True, False])
@pytest.mark.parametrize("yes", [True, False])
def test_clean(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
    venvs_parent_path: Path,
    dry: bool,
    yes: bool,
) -> None:
    paths = venvs_parent_path.glob("is_venv_*")
    clirunner.invoke(
        click_app,
        [
            "link",
            "--workon-home",
            str(workon_home_with_is_venv),
            "-vv",
            *map(str, paths),
        ],
    )

    # additional bad venv
    path = venvs_parent_path / "no_venv_0"
    link = workon_home_with_is_venv / "no_venv_0"

    link.symlink_to(path)

    assert link.exists()
    assert link.readlink() == path

    clirunner.invoke(
        click_app,
        [
            "clean",
            "--workon-home",
            str(workon_home_with_is_venv),
            *(["--yes"] if yes else ["--no"]),
            *(["--dry-run"] if dry else []),
        ],
    )

    if dry:
        assert link.exists()
    elif yes:
        assert not link.exists()
    else:
        assert link.exists()


def test_run_help(
    click_app: Command,
    clirunner: CliRunner,
) -> None:
    out = clirunner.invoke(
        click_app,
        ["run"],
    )

    assert "Run uv commands using" in out.output


@pytest.mark.parametrize("dry", [True])
@pytest.mark.parametrize("named", [True, False])
@pytest.mark.parametrize("resolve", [True, False])
# @pytest.mark.parametrize("resolve", [False])
def test_run(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
    dry: bool,
    named: bool,
    resolve: bool,
) -> None:
    args = ["python", "-c", "import sys; print(sys.executable)"]

    path = workon_home_with_is_venv / "is_venv_0"
    if resolve:
        path = path.resolve()

    opts = (
        ["-n", "is_venv_0"]
        if named
        else ["-p", str(workon_home_with_is_venv / "is_venv_0")]
    )

    out = clirunner.invoke(
        click_app,
        [
            "run",
            "--workon-home",
            str(workon_home_with_is_venv),
            *opts,
            *args,
            *(["--dry-run"] if dry else []),
            *([] if resolve else ["--no-resolve"]),
        ],
    )

    if dry:
        expected = f"VIRTUAL_ENV={path} UV_PROJECT_ENVIRONMENT={path} uv run -p {path} --no-project {shlex.join(args)}"
        assert expected == out.output.strip()


def test_shell_config(click_app: Command, clirunner: CliRunner) -> None:
    out = clirunner.invoke(click_app, ["shell-config"])
    assert out.output.strip() == generate_shell_config().strip()


def test_shell_activate(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
) -> None:
    out = clirunner.invoke(
        click_app,
        ["activate", "-n", "is_venv_0", "--workon-home", str(workon_home_with_is_venv)],
    )
    assert f"source {workon_home_with_is_venv}/is_venv_0" in out.output

    out = clirunner.invoke(
        click_app,
        [
            "activate",
            "-p",
            str(workon_home_with_is_venv / "is_venv_1"),
            "--workon-home",
            str(workon_home_with_is_venv),
        ],
    )
    assert f"source {workon_home_with_is_venv}/is_venv_1" in out.output

    out = clirunner.invoke(
        click_app,
        [
            "activate",
            "-p",
            str(workon_home_with_is_venv / "is_venv_2"),
            "--workon-home",
            str(workon_home_with_is_venv),
        ],
    )

    assert out.exit_code == 1


def test_shell_cd(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
) -> None:
    out = clirunner.invoke(
        click_app,
        ["cd", "-n", "is_venv_0", "--workon-home", str(workon_home_with_is_venv)],
    )

    path = (workon_home_with_is_venv / "is_venv_0").resolve().parent

    assert f"cd {path}" in out.output


def test__main__() -> None:
    from subprocess import check_call

    assert not check_call(["python", "-m", "uv_workon"])


@skip_if_no_jupyter_client
def test_install_ipykernels(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
    venvs_parent_path: Path,
) -> None:
    expected_all = [
        f"ipykernel_install_script.py --dry-run -- --name is_venv_{i} --display-name 'Python [venv: is_venv_{i}]' --user"
        for i in range(3)
    ]

    out = clirunner.invoke(
        click_app,
        [
            "kernels",
            "install",
            "--workon-home",
            str(workon_home_with_is_venv),
            "--all",
            "--dry-run",
        ],
    )
    assert not out.exit_code
    for expected in expected_all:
        assert expected in out.output

    out = clirunner.invoke(
        click_app,
        [
            "kernels",
            "install",
            "--workon-home",
            str(workon_home_with_is_venv),
            "--dry-run",
            "-n",
            "is_venv_0",
        ],
    )
    assert not out.exit_code
    assert expected_all[0] in out.output
    for not_expected in expected_all[1:]:
        assert not_expected not in out.output

    out = clirunner.invoke(
        click_app,
        [
            "kernels",
            "install",
            "--workon-home",
            str(workon_home_with_is_venv),
            "--dry-run",
            "-p",
            str(venvs_parent_path / "has_dotvenv_0"),
        ],
    )

    assert not out.exit_code
    assert (
        "ipykernel_install_script.py --dry-run -- --name has_dotvenv_0 --display-name 'Python [venv: has_dotvenv_0]' --user"
        in out.output
    )


@skip_if_no_jupyter_client
@pytest.mark.parametrize("yes", [True, False])
def test_install_ipykernels_replace(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
    venvs_parent_path: Path,
    dummy_kernelspec_with_replace: dict[str, Any],
    mocker: MockerFixture,
    yes: bool,
) -> None:
    mocker.patch(
        "uv_workon.kernels.get_kernelspecs",
        autospec=True,
        return_value=dummy_kernelspec_with_replace,
    )
    mocker.patch("typer.confirm", autospec=True, return_value=False)
    mocked_uv_run = mocker.patch("uv_workon.cli.uv_run", autospec=True)

    out = clirunner.invoke(
        click_app,
        [
            "kernels",
            "install",
            "--workon-home",
            str(workon_home_with_is_venv),
            "-p",
            str(venvs_parent_path / "is_venv_0"),
            *(["--yes"] if yes else []),
        ],
    )

    assert not out.exit_code

    if yes:
        assert mocked_uv_run.mock_calls == [
            mocker.call(
                (venvs_parent_path / "is_venv_0").resolve(),
                "python",
                get_ipykernel_install_script_path(),
                "--",
                "--name",
                "is_venv_0",
                "--display-name",
                "Python [venv: is_venv_0]",
                "--user",
                dry_run=False,
            )
        ]

    else:
        assert mocked_uv_run.mock_calls == []


@skip_if_no_jupyter_client
@pytest.mark.parametrize("yes", [True, False])
@pytest.mark.parametrize(
    ("args", "expected"),
    [
        (("-n", "dummy0"), ["dummy0"]),
        (("-n", "dummy0", "--dry-run"), []),
        (("-n", "something"), []),
        (("--missing",), ["dummy0", "dummy1"]),
    ],
)
def test_remove_ipykernels(
    click_app: Command,
    clirunner: CliRunner,
    workon_home_with_is_venv: Path,
    mocker: MockerFixture,
    dummy_kernelspec: dict[str, Any],
    yes: bool,
    args: tuple[str],
    expected: list[str],
) -> None:
    mocker.patch(
        "uv_workon.kernels.get_kernelspecs",
        autospec=True,
        return_value=dummy_kernelspec,
    )
    mocker.patch(
        "typer.confirm",
        return_value=False,
    )

    mocked_remove_kernelspecs = mocker.patch("uv_workon.kernels.remove_kernelspecs")

    out = clirunner.invoke(
        click_app,
        [
            "kernels",
            "remove",
            "--workon-home",
            str(workon_home_with_is_venv),
            *args,
            *(["--yes"] if yes else []),
        ],
    )

    assert not out.exit_code

    if yes and expected:
        assert mocked_remove_kernelspecs.mock_calls == [mocker.call(expected)]
    else:
        assert mocked_remove_kernelspecs.mock_calls == []


@skip_if_no_jupyter_client
def test_list_kernels(
    click_app: Command,
    clirunner: CliRunner,
    mocker: MockerFixture,
) -> None:
    mocked = mocker.patch("jupyter_client.kernelspecapp.ListKernelSpecs")

    clirunner.invoke(
        click_app,
        [
            "kernels",
            "list",
        ],
    )

    assert mocked.mock_calls == [mocker.call(log_level="ERROR"), mocker.call().start()]



from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

import pytest

from uv_workon import kernels

if TYPE_CHECKING:
    from typing import Any

    from pytest_mock import MockerFixture

_has_jupyter_client = find_spec("jupyter_client") is not None

skip_if_no_jupyter_client = pytest.mark.skipif(
    not _has_jupyter_client, reason="Must install jupyter client for this test."
)


def test_get_ipykernel_install_script_path() -> None:
    from importlib.resources import files

    assert (
        str(
            files("uv_workon")
            .joinpath("scripts")
            .joinpath("ipykernel_install_script.py")
        )
        == kernels.get_ipykernel_install_script_path()
    )


# dumb tests.  to be replaced with better someday...
def test_has_jupyter_client() -> None:
    if _has_jupyter_client:
        kernels.has_jupyter_client()
    else:
        with pytest.raises(ModuleNotFoundError):
            kernels.has_jupyter_client()


@skip_if_no_jupyter_client
def test_get_kernelspec_0() -> None:
    assert isinstance(kernels.get_kernelspecs(), dict)
    assert isinstance(kernels.get_broken_kernelspecs(), dict)
    kernels.remove_kernelspecs([])


@skip_if_no_jupyter_client
def test_get_broken_kernelspecs(
    mocker: MockerFixture,
    dummy_kernelspec: dict[str, Any],
    mocked_get_kernelspec: Any,
) -> None:
    assert kernels.get_broken_kernelspecs() == {
        k: v for k, v in dummy_kernelspec.items() if k != "good"
    }
    assert mocked_get_kernelspec.mock_calls == [mocker.call()]


@skip_if_no_jupyter_client
def test_complete_kernelspec_names(
    mocker: MockerFixture,
    mocked_get_kernelspec: Any,
) -> None:
    names = ["dummy0", "dummy1", "good"]

    assert list(kernels.complete_kernelspec_names("")) == names
    assert list(kernels.complete_kernelspec_names("d")) == names[:-1]
    assert list(kernels.complete_kernelspec_names("g")) == names[-1:]

    assert mocked_get_kernelspec.mock_calls == [mocker.call()] * 3


@skip_if_no_jupyter_client
def test_remove_kernelspecs(mocker: MockerFixture) -> None:
    mock_removekernelspec = mocker.patch(
        "jupyter_client.kernelspecapp.RemoveKernelSpec", autospec=True
    )
    kernels.remove_kernelspecs(["a", "b"])
    assert mock_removekernelspec.mock_calls == [
        mocker.call(spec_names=["a", "b"], force=True),
        mocker.call().start(),
    ]



# pylint: disable=wrong-import-position
"""Config file for nox."""

# * Imports ----------------------------------------------------------------------------
from __future__ import annotations

import shlex
import shutil
import sys
from dataclasses import dataclass
from functools import lru_cache, partial, wraps
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Literal,
    TypeAlias,
    TypedDict,
)

import nox
from nox.virtualenv import CondaEnv

sys.path.insert(0, ".")
from tools.dataclass_parser import (
    DataclassParser,
    add_option,
    option,
)
from tools.noxtools import (
    check_for_change_manager,
    combine_list_str,
    get_python_full_path,
    infer_requirement_path,
    open_webpage,
    session_run_commands,
)

sys.path.pop(0)


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Sequence
    from os import PathLike
    from typing import Any

    from nox import Session


# * Names ------------------------------------------------------------------------------

PACKAGE_NAME = "uv-workon"
IMPORT_NAME = "uv_workon"
KERNEL_NAME = "uv-workon"

# * nox options ------------------------------------------------------------------------

ROOT = Path(__file__).parent

nox.needs_version = ">=2024.10.9"
nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ["lint", "typing", "test-all"]
nox.options.default_venv_backend = "uv"

# * Options ---------------------------------------------------------------------------

# if True, use uv lock/sync.  If False, use uv pip compile/sync...
UV_LOCK = True

PYTHON_ALL_VERSIONS = [
    c.split()[-1]
    for c in nox.project.load_toml("pyproject.toml")["project"]["classifiers"]
    if c.startswith("Programming Language :: Python :: 3.")
]
PYTHON_DEFAULT_VERSION = Path(".python-version").read_text(encoding="utf-8").strip()

UVX_LOCK_CONSTRAINTS = "requirements/lock/uvx-tools.txt"
UVX_MIN_CONSTRAINTS = "requirements/uvx-tools.txt"
PIP_COMPILE_CONFIG = "requirements/uv.toml"


class SessionOptionsDict(TypedDict, total=False):
    """Dict for options to nox.session."""

    python: str | list[str]
    venv_backend: str | Callable[..., CondaEnv]


CONDA_DEFAULT_KWS: SessionOptionsDict = {
    "python": PYTHON_DEFAULT_VERSION,
    "venv_backend": "micromamba|mamba|conda",
}
CONDA_ALL_KWS: SessionOptionsDict = {
    "python": PYTHON_ALL_VERSIONS,
    "venv_backend": "micromamba|mamba|conda",
}

DEFAULT_KWS: SessionOptionsDict = {"python": PYTHON_DEFAULT_VERSION}
ALL_KWS: SessionOptionsDict = {"python": PYTHON_ALL_VERSIONS}


# * Session command line options -------------------------------------------------------

OPT_TYPE: TypeAlias = list[str] | None
RUN_TYPE: TypeAlias = list[list[str]] | None

RUN_ANNO = Annotated[
    RUN_TYPE,
    option(
        help="Run external commands in session.  Specify multiple times for multiple commands.",
    ),
]
OPT_ANNO = Annotated[OPT_TYPE, option(help="Options to command.")]


@dataclass
class SessionParams(DataclassParser):
    """Holds all cli options."""

    # common parameters
    lock: bool = False
    no_lock: bool = False
    update: bool = add_option("--update", "-U", help="update dependencies/package")
    version: str | None = add_option(
        "--version", "-V", help="pretend version", default=None
    )
    prune: bool = add_option(default=False, help="Pass `--prune` to conda env update")
    no_frozen: bool = add_option(
        "--no-frozen",
        "-N",
        help="run `uv sync` without --frozen (default is to use `--frozen`)",
    )
    reinstall_package: bool = add_option(
        "--reinstall-package",
        "-P",
        help="reinstall package.  Only works with uv sync and editable installs",
    )
    installpkg: str | None = add_option(
        "--installpkg",
        help="Use this package instead of editable or built package",
        default=None,
    )

    # requirements
    requirements_no_notify: bool = add_option(
        default=False,
        help="Skip notification of lock-compile",
    )

    # lock
    lock_force: bool = False
    lock_upgrade: bool = add_option(
        "--lock-upgrade",
        "-L",
        help="Upgrade all packages in lock files",
        default=False,
    )

    # test
    test_no_pytest: bool = False
    test_options: OPT_TYPE = add_option(
        "--test-options", "-t", help="Options to pytest"
    )
    test_run: RUN_ANNO = None
    no_cov: bool = False

    # coverage
    coverage: list[Literal["erase", "combine", "report", "html", "open"]] | None = None

    # docs
    docs: (
        list[
            Literal[
                "html",
                "build",
                "symlink",
                "clean",
                "livehtml",
                "linkcheck",
                "spelling",
                "showlinks",
                "open",
                "serve",
            ]
        ]
        | None
    ) = add_option("--docs", "-d", help="doc commands")
    docs_run: RUN_ANNO = None
    docs_options: OPT_TYPE = add_option(
        "--docs-options", help="Options to sphinx-build"
    )
    # lint
    lint_options: OPT_TYPE = add_option(help="Options to pre-commit")

    # typing
    typing: list[
        Literal[
            "clean",
            "mypy",
            "pyright",
            "pylint",
            "pytype",
            "all",
            "notebook-mypy",
            "notebook-pyright",
            "notebook-typecheck",
        ]
    ] = add_option("--typing", "-m")
    typing_run: RUN_ANNO = None
    typing_options: OPT_TYPE = add_option(help="Options to type checkers")

    # build
    build: list[Literal["build", "version"]] | None = None
    build_run: RUN_ANNO = None
    build_isolation: bool = False
    build_out_dir: str = "./dist"
    build_options: OPT_ANNO = None
    build_silent: bool = False

    # publish
    publish: list[Literal["release", "test", "check"]] | None = add_option(
        "-p", "--publish"
    )

    # conda-recipe/grayskull
    conda_recipe: list[Literal["recipe", "recipe-full"]] | None = None
    conda_recipe_sdist_path: str | None = None

    # conda-build
    conda_build: list[Literal["build", "clean"]] | None = None
    conda_build_run: RUN_ANNO = None


@lru_cache
def parse_posargs(*posargs: str) -> SessionParams:
    """
    Get Parser using `+` for session option prefix.

    Note that using `+` allows for passing underlying `-` options
    without escaping.
    """
    opts = SessionParams.from_posargs(posargs=posargs, prefix_char="+")

    if opts.no_lock:
        opts.lock = False
    else:
        opts.lock = opts.lock or UV_LOCK

    return opts


def add_opts(
    func: Callable[[Session, SessionParams], None],
) -> Callable[[Session], None]:
    """Fill in `opts` from cli options."""

    @wraps(func)
    def wrapped(session: Session) -> None:
        opts = parse_posargs(*session.posargs)
        return func(session, opts)

    return wrapped


# * Dependencies --------------------------------------------------------------
def install_dependencies(
    session: Session,
    *args: str,
    name: str,
    opts: SessionParams,
    python_version: str | None = None,
    location: str | None = None,
    no_dev: bool = True,
    only_group: bool = False,
    include_editable_package: bool = False,
    lock: bool | None = None,
) -> None:
    """General dependencies installer."""
    if python_version is None:
        assert isinstance(session.python, str)  # noqa: S101
        python_version = session.python

    lock = lock if lock is not None else opts.lock

    if isinstance(session.virtualenv, CondaEnv):
        environment_file = infer_requirement_path(
            name,
            ext=".yaml",
            python_version=python_version,
            lock=False,
        )
        with check_for_change_manager(
            environment_file,
            hash_path=Path(session.create_tmp()) / "env.json",
        ) as changed:
            if changed or opts.update:
                session.run_install(
                    session.virtualenv.conda_cmd,
                    "env",
                    "update",
                    "--yes",
                    *(["--prune"] if opts.prune else []),
                    "-f",
                    environment_file,
                    "--prefix",
                    session.virtualenv.location,
                    *args,
                )
            else:
                session.log("Using cached install")

        if include_editable_package:
            install_package(session, editable=True, update=True)

    elif lock:  # pylint: disable=confusing-consecutive-elif
        session.run_install(
            "uv",
            "sync",
            *(["-U"] if opts.update else []),
            *(["--no-dev"] if no_dev else []),
            *([] if opts.no_frozen else ["--frozen"]),
            *(["--only-group"] if only_group else ["--group"]),
            name,
            # Handle package install here?
            # "--no-editable",
            # "--reinstall-package",
            # "open-notebook",
            *([] if include_editable_package else ["--no-install-project"]),
            *(
                [f"--reinstall-package={PACKAGE_NAME}"]
                if opts.reinstall_package and include_editable_package
                else []
            ),
            *(
                []
                if any("--python" in a for a in args)
                else [f"--python={python_version}"]
            ),
            *args,
            env={"UV_PROJECT_ENVIRONMENT": location or session.virtualenv.location},
        )

    else:
        session.run_install(
            "uv",
            "pip",
            "sync",
            f"--config-file={PIP_COMPILE_CONFIG}",
            infer_requirement_path(
                name,
                ext=".txt",
                python_version=python_version,
                lock=True,
            ),
            *args,
        )

        if include_editable_package:
            install_package(session, editable=True, update=True)


def install_package(
    session: Session,
    *args: str,
    editable: bool = False,
    update: bool = True,
    installpkg: str | None = None,
) -> None:
    """Install current package."""
    if installpkg is not None:
        run = session.run
        opts = [*args, installpkg]
    elif editable:
        run = session.run if update else session.run_install
        opts = [*args, "-e", "."]
    else:
        run = session.run
        opts = [*args, get_package_wheel(session)]

    run(
        "uv",
        "pip",
        "install",
        *opts,
        "--no-deps",
        "--force-reinstall",
        external=True,
    )


def get_package_wheel(
    session: Session,
    opts: str | Iterable[str] | None = None,
    extras: str | Iterable[str] | None = None,
    reuse: bool = True,
) -> str:
    """
    Build the package in return the build location.

    This is similar to how tox does isolated builds.

    Note that the first time this is called,

    Should be straightforward to extend this to isolated builds
    that depend on python version (something like have session build-3.11 ....)
    """
    dist_location = Path(session.cache_dir) / "dist"
    if reuse and getattr(get_package_wheel, "_called", False):
        session.log("Reuse isolated build")
    else:
        cmd = f"nox -s build -- ++build-out-dir {dist_location} ++build-options --wheel ++build-silent"
        session.run_always(*shlex.split(cmd), external=True)

        # save that this was called:
        if reuse:
            get_package_wheel._called = True  # type: ignore[attr-defined]  # noqa: SLF001  # pylint: disable=protected-access

    paths = list(dist_location.glob("*.whl"))
    if len(paths) != 1:
        msg = f"something wonky with paths {paths}"
        raise ValueError(msg)

    path = f"{PACKAGE_NAME}@{paths[0]}"
    if extras:
        if not isinstance(extras, str):
            extras = ",".join(extras)
        path = f"{path}[{extras}]"

    if opts:
        if not isinstance(opts, str):
            opts = " ".join(opts)
        path = f"{path} {opts}"

    return path


# * uvx runner ----------------------------------------------------------------
def get_uvx_constraint_args(locked: bool = True) -> tuple[str, ...]:
    """Get constraints file for uvx."""
    if locked and Path(UVX_LOCK_CONSTRAINTS).exists():
        return (f"--constraints={UVX_LOCK_CONSTRAINTS}",)
    if Path(UVX_MIN_CONSTRAINTS).exists():
        return (f"--constraints={UVX_MIN_CONSTRAINTS}",)
    return ()


def uvx_run(
    session: Session, *args: str | PathLike[str], locked: bool = True, **kwargs: Any
) -> Any:
    """Run command using uvx"""
    return session.run("uvx", *get_uvx_constraint_args(locked), *args, **kwargs)


def pre_commit_run(session: Session, *args: str | PathLike[str], **kwargs: Any) -> Any:
    """Run pre-commit via uvx."""
    return uvx_run(
        session,
        "--with=pre-commit-uv",
        "pre-commit",
        "run",
        *args,
        **kwargs,
        locked=False,
    )


# * Sessions ------------------------------------------------------------------
# ** test-all
@nox.session(name="test-all", python=False)
def test_all(session: Session) -> None:
    """Run all tests and coverage."""
    session.notify("coverage-erase")
    for py in PYTHON_ALL_VERSIONS:
        session.notify(f"test-{py}")
    session.notify("test-noopt")
    session.notify("coverage")


# ** dev
@nox.session(name="dev", python=False)
@add_opts
def dev(
    session: Session,
    opts: SessionParams,
) -> None:
    """Create development environment."""
    session.run("uv", "venv", ".venv", "--allow-existing", "--prompt", PACKAGE_NAME)

    python_opt = "--python=.venv/bin/python"

    install_dependencies(
        session,
        python_opt,
        name="dev",
        opts=opts,
        python_version=PYTHON_DEFAULT_VERSION,
        location=".venv",
        no_dev=False,
        include_editable_package=True,
    )
    session.notify("install-ipykernel")


@nox.session(name="install-ipykernel", python=False)
def install_ipykernel(session: Session) -> None:
    """Install ipykernel for .venv"""
    session.run(
        "uv",
        "run",
        "--frozen",
        "--python=.venv/bin/python",
        "python",
        "-m",
        "ipykernel",
        "install",
        "--user",
        "--name",
        KERNEL_NAME,
        "--display-name",
        f"Python [venv: {KERNEL_NAME}]",
        success_codes=[0, 1],
    )


# ** requirements
@nox.session(name="requirements", python=False)
@add_opts
def requirements(
    session: Session,
    opts: SessionParams,
) -> None:
    """
    Create environment.yaml and requirement.txt files from pyproject.toml using pyproject2conda.

    These will be placed in the directory "./requirements".

    Should instead us pre-commit run requirements --all-files
    """
    pre_commit_run(
        session,
        "pyproject2conda-project",
        "--all-files",
        success_codes=[0, 1],
    )

    if not opts.requirements_no_notify:
        session.notify("lock")


# ** uv lock compile
@nox.session(name="lock", python=False)
@add_opts
def lock(
    session: Session,
    opts: SessionParams,
) -> None:
    """Run uv pip compile ..."""
    options: list[str] = ["-U"] if opts.lock_upgrade else []
    force = opts.lock_force or opts.lock_upgrade

    if opts.lock and opts.lock_upgrade:
        session.run(
            "uv",
            "sync" if opts.update else "lock",
            "--upgrade",
            env={
                "VIRTUAL_ENV": ".venv",
                "UV_PROJECT_ENVIRONMENT": ".venv",
            },
        )

    from packaging.version import Version

    min_python_version = min(PYTHON_ALL_VERSIONS, key=Version)

    reqs_path = Path("./requirements")
    for path in reqs_path.glob("*.txt"):
        python_version = (
            min_python_version
            if path.name
            in {"test.txt", "test-extras.txt", "typing.txt", "uvx-tools.txt"}
            else PYTHON_DEFAULT_VERSION
        )

        lockpath = infer_requirement_path(
            path.name,
            ext=".txt",
            python_version=python_version,
            lock=True,
            check_exists=False,
        )

        with check_for_change_manager(
            path,
            target_path=lockpath,
            force_write=force,
        ) as changed:
            if force or changed:
                session.run(
                    "uv",
                    "pip",
                    "compile",
                    "--universal",
                    f"--config-file={PIP_COMPILE_CONFIG}",
                    "-q",
                    # don't include dependencies for uvx-tools
                    *(["--no-deps"] if path.name == "uvx-tools.txt" else []),
                    "--python-version",
                    python_version,
                    *options,
                    path,
                    "-o",
                    lockpath,
                )
            else:
                session.log(f"Skipping {lockpath}")


# ** testing
def _test(
    session: nox.Session,
    run: RUN_TYPE,
    test_no_pytest: bool,
    test_options: OPT_TYPE,
    no_cov: bool,
) -> None:
    import os

    tmpdir = os.environ.get("TMPDIR", None)

    session_run_commands(session, run)
    if not test_no_pytest:
        opts = combine_list_str(test_options or [])
        if not no_cov:
            session.env["COVERAGE_FILE"] = str(Path(session.create_tmp()) / ".coverage")

            if not any(o.startswith("--cov") for o in opts):
                opts.append(f"--cov={IMPORT_NAME}")

        # Because we are testing if temporary folders
        # have git or not, we have to make sure we're above the
        # not under this repo
        # so revert to using the top level `TMPDIR`
        if tmpdir:
            session.env["TMPDIR"] = tmpdir

        session.run("pytest", *opts)


# *** Basic tests
@add_opts
def test(
    session: Session,
    opts: SessionParams,
) -> None:
    """Test environments with conda installs."""
    install_dependencies(session, name="test", opts=opts)
    install_package(session, editable=False, update=True, installpkg=opts.installpkg)

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_options=opts.test_options,
        no_cov=opts.no_cov,
    )


nox.session(**ALL_KWS)(test)
nox.session(name="test-conda", **CONDA_ALL_KWS)(test)


@nox.session(name="test-noopt", **DEFAULT_KWS)
@add_opts
def test_noopt(
    session: Session,
    opts: SessionParams,
) -> None:
    """Test environments with conda installs."""
    install_dependencies(session, name="test-noopt", opts=opts)
    install_package(session, editable=False, update=True)

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_options=opts.test_options,
        no_cov=opts.no_cov,
    )


@nox.session(name="test-notebook", **DEFAULT_KWS)
@add_opts
def test_notebook(session: nox.Session, opts: SessionParams) -> None:
    """Run pytest --nbval."""
    install_dependencies(session, name="test-notebook", opts=opts)
    install_package(session, editable=False, update=True, installpkg=opts.installpkg)

    test_nbval_opts = shlex.split(
        """
    --nbval
    --nbval-current-env
    --nbval-sanitize-with=config/nbval.ini
    --dist loadscope
   """,
    )

    test_options = (
        (opts.test_options or [])
        + test_nbval_opts
        + [str(p) for p in Path("examples/usage").glob("*.ipynb")]
    )

    session.log(f"{test_options = }")

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_options=test_options,
        no_cov=opts.no_cov,
    )


@nox.session(python=False)
@add_opts
def coverage(
    session: Session,
    opts: SessionParams,
) -> None:
    """Run coverage."""
    cmd = opts.coverage or ["combine", "html", "report"]

    paths = list(Path(".nox").glob("test-*/tmp/.coverage*"))

    if "erase" in cmd:
        for path in paths:
            if path.exists():
                session.log(f"removing {path}")
                path.unlink()

    for c in cmd:
        if c == "combine":
            uvx_run(
                session,
                "coverage",
                "combine",
                "--keep",
                "-a",
                *paths,
            )
        elif c == "open":
            open_webpage(path="htmlcov/index.html")

        else:
            uvx_run(
                session,
                "coverage",
                c,
            )


@nox.session(name="coverage-erase", python=False)
def coverage_erase(session: Session) -> None:
    """Alias to `nox -s coverage -- ++coverage erase`"""
    session.run("nox", "-s", "coverage", "--", "++coverage", "erase")


# *** testdist (conda)
@add_opts
def testdist(
    session: Session,
    opts: SessionParams,
) -> None:
    """Test conda distribution."""
    install_str = PACKAGE_NAME
    if opts.version:
        install_str = f"{install_str}=={opts.version}"

    install_dependencies(session, name="test-extras", only_group=True, opts=opts)

    if isinstance(session.virtualenv, CondaEnv):
        session.conda_install(install_str)
    else:
        session.install(install_str)

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_options=opts.test_options,
        no_cov=opts.no_cov,
    )


nox.session(name="testdist-pypi", **ALL_KWS)(testdist)
nox.session(name="testdist-conda", **CONDA_ALL_KWS)(testdist)


# # ** Docs
@nox.session(name="docs", **DEFAULT_KWS)
@add_opts
def docs(  # noqa: C901, PLR0912
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """
    Run `make` in docs directory.

    For example, 'nox -s docs -- +d html'
    calls 'make -C docs html'. With 'release' option, you can set the
    message with 'message=...' in posargs.
    """
    cmd = opts.docs or []
    cmd = ["html"] if not opts.docs_run and not cmd else list(cmd)
    name = "docs-live" if "livehtml" in cmd else "docs"

    install_dependencies(session, name=name, opts=opts, include_editable_package=True)

    if opts.version:
        session.env["SETUPTOOLS_SCM_PRETEND_VERSION"] = opts.version
    session_run_commands(session, opts.docs_run)

    if "symlink" in cmd:
        cmd.remove("symlink")
        _create_doc_examples_symlinks(session)

    if open_page := "open" in cmd:
        cmd.remove("open")

    if serve := "serve" in cmd:
        open_webpage(url="http://localhost:8000")
        cmd.remove("serve")

    if cmd:
        common_opts = [
            "--doctree-dir=docs/_build/doctree",
            *(opts.docs_options or ()),
        ]
        for c in combine_list_str(cmd):
            if c == "clean":
                for d in ("docs/_build", "generated", "reference/generated"):
                    shutil.rmtree(Path(d), ignore_errors=True)
                session.log("cleaned docs")
            elif c == "livehtml":
                session.run(
                    "sphinx-autobuild",
                    "-b",
                    "html",
                    "docs",
                    "docs/_build/html",
                    *common_opts,
                    "--open-browser",
                    *(
                        f"--ignore='*/{d}/*'"
                        for d in (
                            "_build",
                            "generated",
                            "jupyter_execute",
                            ".ipynb_checkpoints",
                        )
                    ),
                )
            elif c == "showlinks":
                session.run(
                    "python",
                    "-m",
                    "sphinx.ext.intersphinx",
                    "docs/_build/html/objects.inv",
                )
            else:
                session.run(
                    "sphinx-build", "-b", c, *common_opts, "docs", f"docs/_build/{c}"
                )

    if open_page:
        open_webpage(path="./docs/_build/html/index.html")

    if serve and "livehtml" not in cmd:
        session.run(
            "python",
            "-m",
            "http.server",
            "-d",
            "docs/_build/html",
            "-b",
            "127.0.0.1",
            "8000",
        )


# ** lint
@nox.session(python=False)
@add_opts
def lint(
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """
    Run linters with pre-commit.

    Defaults to `pre-commit run --all-files`.
    To run something else pass, e.g.,
    `nox -s lint -- --lint-run "pre-commit run --hook-stage manual --all-files`
    """
    pre_commit_run(session, "--all-files", *(opts.lint_options or []))


# ** type checking
@nox.session(name="typing", **ALL_KWS)
@add_opts
def typing(  # noqa: C901
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """Run type checkers (mypy, pyright, pytype)."""
    install_dependencies(
        session, name="typing", opts=opts, include_editable_package=True
    )
    session_run_commands(session, opts.typing_run)

    cmd = opts.typing or []
    if not opts.typing_run and not cmd:
        cmd = ["mypy", "pyright", "pylint"]

    if "all" in cmd:
        cmd = ["mypy", "pyright", "pylint", "pytype"]

    # set the cache directory for mypy
    session.env["MYPY_CACHE_DIR"] = str(Path(session.create_tmp()) / ".mypy_cache")

    if "clean" in cmd:
        cmd = list(cmd)
        cmd.remove("clean")

        for name in (".mypy_cache", ".pytype"):
            p = Path(session.create_tmp()) / name
            if p.exists():
                session.log(f"removing cache {p}")
                shutil.rmtree(p)

    if not isinstance(session.python, str):
        raise TypeError

    for c in cmd:
        if c.startswith("notebook-"):
            session.run("make", c, external=True)
        elif c in {"mypy", "pyright"}:
            session.run(
                "python",
                "tools/typecheck.py",
                *get_uvx_constraint_args(),
                "--verbose",
                f"--checker={c}",
                "--",
                *(opts.typing_options or []),
                *(["--color-output"] if c == "mypy" else []),
            )
        elif c == "pylint":
            session.run(
                "pylint",
                # A bit dangerous, but needed to allow pylint
                # to work across versions.
                "--disable=unrecognized-option",
                "--enable-all-extensions",
                "src",
                "tests",
            )
        else:
            session.log(f"Skipping unknown command {c}")


# ** Dist pypi
# NOTE: you can skip having the build environment and
# just use uv build, but faster to use environment ...
USE_ENVIRONMENT_FOR_BUILD = False
_build_dec = nox.session(
    python=PYTHON_DEFAULT_VERSION if USE_ENVIRONMENT_FOR_BUILD else False
)


@_build_dec
@add_opts
def build(session: nox.Session, opts: SessionParams) -> None:  # noqa: C901
    """
    Build the distribution.

    Note that default is to not use build isolation.
    Pass `--build-isolation` to use build isolation.
    """
    if USE_ENVIRONMENT_FOR_BUILD:
        install_dependencies(session, name="build", opts=opts, lock=False)

    if opts.version:
        session.env["SETUPTOOLS_SCM_PRETEND_VERSION"] = opts.version

    for cmd in opts.build or ["build"]:
        if cmd == "version":
            if USE_ENVIRONMENT_FOR_BUILD:
                session.run(get_python_full_path(session), "-m", "hatchling", "version")  # pyright: ignore[reportPossiblyUnboundVariable]
            else:
                uvx_run(
                    session, "--with=hatch-vcs", "hatchling", "version", external=True
                )
        elif cmd == "build":
            outdir = opts.build_out_dir
            shutil.rmtree(outdir, ignore_errors=True)

            args = shlex.split(f"uv build --out-dir={outdir}")
            if USE_ENVIRONMENT_FOR_BUILD and not opts.build_isolation:
                args.append("--no-build-isolation")

            if opts.build_options:
                args.extend(opts.build_options)

            out = session.run(*args, silent=opts.build_silent)
            if opts.build_silent:
                if not isinstance(out, str):
                    msg = "session.run output not a string"
                    raise ValueError(msg)
                session.log(out.strip().split("\n")[-1])


@nox.session(python=False)
@add_opts
def publish(session: nox.Session, opts: SessionParams) -> None:
    """Publish the distribution."""
    for cmd in opts.publish or []:
        if cmd == "test":
            uvx_run(session, "twine", "upload", "--repository", "testpypi", "dist/*")
        elif cmd == "release":
            uvx_run(session, "twine", "upload", "dist/*")
        elif cmd == "check":
            uvx_run(session, "twine", "check", "--strict", "dist/*")


# # ** Dist conda
@nox.session(name="conda-recipe", python=False)
@add_opts
def conda_recipe(
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """Run grayskull to create recipe."""
    commands = opts.conda_recipe or ["recipe"]

    run = partial(uvx_run, session, external=True)

    if not (sdist_path := opts.conda_recipe_sdist_path):
        sdist_path = PACKAGE_NAME
        if opts.version:
            sdist_path = f"{sdist_path}=={opts.version}"

    for command in commands:
        if command == "recipe":
            # make directory?
            if not (d := Path("./dist-conda")).exists():
                d.mkdir()

            run(
                "grayskull",
                "pypi",
                sdist_path,
                "--sections",
                "package",
                "source",
                "build",
                "requirements",
                "-o",
                "dist-conda",
            )

            _append_recipe(
                f"dist-conda/{PACKAGE_NAME}/meta.yaml",
                "config/recipe-append.yaml",
            )
            session.run("cat", f"dist-conda/{PACKAGE_NAME}/meta.yaml", external=True)

        elif command == "recipe-full":
            import tempfile

            with tempfile.TemporaryDirectory() as d:  # type: ignore[assignment,unused-ignore]
                run(
                    "grayskull",
                    "pypi",
                    sdist_path,
                    "-o",
                    str(d),
                )
                path = Path(d) / PACKAGE_NAME / "meta.yaml"
                session.log(f"cat {path}:")
                with path.open() as f:
                    for line in f:
                        print(line, end="")  # noqa: T201


@nox.session(name="conda-build", **CONDA_DEFAULT_KWS)
@add_opts
def conda_build(session: nox.Session, opts: SessionParams) -> None:
    """Run `conda mambabuild`."""
    session.conda_install("boa", "anaconda-client")
    cmds, run = opts.conda_build, opts.conda_build_run

    session_run_commands(session, run)

    if not run and not cmds:
        cmds = ["build", "clean"]

    if cmds is None:
        cmds = []

    cmds = list(cmds)
    if "clean" in cmds:
        cmds.remove("clean")
        session.log("removing directory dist-conda/build")
        shutil.rmtree(Path("./dist-conda/build"), ignore_errors=True)

    for cmd in cmds:
        if cmd == "build":
            if not (d := Path(f"./dist-conda/{PACKAGE_NAME}/meta.yaml")).exists():
                msg = f"no file {d}"
                raise ValueError(msg)

            session.run(
                "conda",
                "mambabuild",
                "--output-folder=dist-conda/build",
                "--no-anaconda-upload",
                "dist-conda",
            )


# ** Other utilities
@nox.session(**DEFAULT_KWS)
@add_opts
def cog(session: nox.Session, opts: SessionParams) -> None:
    """Run cog."""
    install_dependencies(session, name="cog", opts=opts, include_editable_package=True)
    session.run("cog", "-rP", "README.md", env={"COLUMNS": "90"})


# * Utilities -------------------------------------------------------------------------
def _create_doc_examples_symlinks(session: nox.Session, clean: bool = True) -> None:  # noqa: C901
    """Create symlinks from docs/examples/*.md files to /examples/usage/..."""
    import os

    def usage_paths(path: Path) -> Iterator[Path]:
        with path.open("r") as f:
            for line in f:
                if line.startswith("usage/"):
                    yield Path(line.strip())

    def get_target_path(
        usage_path: str | Path,
        prefix_dir: str | Path = "./examples",
        exts: Sequence[str] = (".md", ".ipynb"),
    ) -> Path:
        path = Path(prefix_dir) / Path(usage_path)

        if not all(ext.startswith(".") for ext in exts):
            msg = "Bad extensions.  Should start with '.'"
            raise ValueError(msg)

        if path.exists():
            return path

        for ext in exts:
            p = path.with_suffix(ext)
            if p.exists():
                return p

        msg = f"no path found for base {path}"
        raise ValueError(msg)

    root = Path("./docs/examples/")
    if clean:
        shutil.rmtree(root / "usage", ignore_errors=True)

    # get all md files
    paths = list(root.glob("*.md"))

    # read usage lines
    for path in paths:
        for usage_path in usage_paths(path):
            target = get_target_path(usage_path)
            link = root / usage_path.parent / target.name

            if link.exists():
                link.unlink()

            link.parent.mkdir(parents=True, exist_ok=True)

            target_rel = os.path.relpath(target, start=link.parent)
            session.log(f"linking {target_rel} -> {link}")

            os.symlink(target_rel, link)


def _append_recipe(recipe_path: str | Path, append_path: str | Path) -> None:
    recipe_path = Path(recipe_path)
    append_path = Path(append_path)

    with recipe_path.open() as f:
        recipe = f.readlines()

    with append_path.open() as f:
        append = f.readlines()

    with recipe_path.open("w") as f:
        f.writelines([*recipe, "\n", *append])



from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    import os
    from collections.abc import Iterable


VirtualEnvPattern: TypeAlias = "str | Iterable[str] | None"
PathLike: TypeAlias = "str | os.PathLike[str]"
